"""OSSOS VOSpace storage convenience package"""
import cStringIO
import errno
import fnmatch
from glob import glob
import os
import re
from string import upper
import tempfile
import logging
import warnings
from astropy.coordinates import SkyCoord
from astropy.io import ascii
from astropy import units
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
import vospace
from astropy.io import fits
import requests
from .downloads.cutouts.calculator import CoordinateConverter
import coding
import util
from astropy.time import Time
from .gui import logger
from .wcs import WCS

# Try and turn off warnings, only works for some releases of requests.
try:
    requests.packages.urllib3.disable_warnings()
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.ERROR)
except:
    pass

MAXCOUNT = 30000
_TARGET = "TARGET"

DBIMAGES = 'vos:OSSOS/dbimages'
MEASURE3 = 'vos:OSSOS/measure3'
POSTAGE_STAMPS = 'vos:OSSOS/postage_stamps'
TRIPLETS = 'vos:OSSOS/triplets'
ASTROM_RELEASES = 'vos:OSSOS/0_OSSOSreleases'

DATA_WEB_SERVICE = 'https://www.canfar.phys.uvic.ca/data/pub/'
VOSPACE_WEB_SERVICE = 'https://www.canfar.phys.uvic.ca/vospace/nodes/'
TAP_WEB_SERVICE = 'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/tap/sync'

OSSOS_TAG_URI_BASE = 'ivo://canfar.uvic.ca/ossos'
OBJECT_COUNT = "object_count"


class Wrapper(object):

    def __init__(self, wrapped_class, *args, **kargs):
        self.wrapped_class = wrapped_class(*args, **kargs)

    def __getattr__(self, attr):
        orig_attr = self.wrapped_class.__getattribute__(attr)
        if callable(orig_attr):
            def hooked(*args, **kwargs):
                print "-->", orig_attr, args, kwargs
                result = orig_attr(*args, **kwargs)
                print "<--", type(result)
                return result
            return hooked
        else:
            return orig_attr

SUCCESS = 'success'

# cache holders.
mopheaders = {}
astheaders = {}
sgheaders = {}
fwhm = {}
zmag = {}
tags = {}

APCOR_EXT = "apcor"
ZEROPOINT_USED_EXT = "zeropoint.used"
PSF_EXT = "psf.fits"


def get_apcor(expnum, ccd, version='p', prefix=None):
    """
    retrieve the aperture correction for this exposure
    @param expnum:
    @param ccd:
    @param version:
    @param prefix:
    @return:
    """
    uri = get_uri(expnum, ccd, ext=APCOR_EXT, version=version, prefix=prefix)
    url = "http://{}/vospace/auth/nodes/{}".format(vospace.VOSPACE_SERVER,
                                                    uri[4:])
    resp = requests.get(url, params={'view': 'data'}, auth=vospace.authentication)
    return [float (x) for x in resp.content.split()]

def get_detections(release='current'):
    """

    @param release: the release to retrieve from VOSpace
    @return: astropy.table.Table
    """
    detection_path = "{}/{}/OSSOS*.detections".format(ASTROM_RELEASES, release)
    filenames = vospace.client.glob(detection_path)
    if not len(filenames) > 0:
        raise IOError("No detection file found using: {}".format(detection_path))

    return ascii.read(open_vos_or_local(filenames[0]).read(), header_start=-1)


def cone_search(ra, dec, dra=0.01, ddec=0.01, mjdate=None, calibration_level=2):
    """Do a QUERY on the TAP service for all observations that are part of OSSOS (*P05/*P016)
    where taken after mjd and have calibration 'observable'.

    @param ra: RA center of search cont
    @type ra: Quantity
    @param dec: float degrees
    @type dec: Quantity
    @param dra: float degrees
    @type dra: Quantity
    @param ddec: float degrees
    @type ddec: Quantity
    @param calibration_level: What calibration level must the found image have,
    @param mjdate: what data must the observation be take on.
    this is a CAOM2 parameter of CADC, 2 means calibrated data.
    """

    data = dict(QUERY=(" SELECT Observation.observationID as collectionID, "
                       " Plane.time_bounds_cval1 AS mjdate, "
                       " Plane.time_exposure AS exptime, "
                       " Plane.energy_bandpassName as filter"
                       " FROM caom2.Observation AS Observation "
                       " JOIN caom2.Plane AS Plane "
                       " ON Observation.obsID = Plane.obsID "
                       " WHERE  ( Observation.collection = 'CFHT' ) "
                       " AND Plane.calibrationLevel={} "
                       " AND ( Plane.energy_bandpassName LIKE 'r.%' OR Plane.energy_bandpassName LIKE 'gri.%' ) "
                       " AND ( Observation.proposal_id LIKE '%P05' or Observation.proposal_id LIKE '%P06' )"
                       " AND Observation.target_name NOT LIKE 'WP%'"),
                REQUEST="doQuery",
                LANG="ADQL",
                FORMAT="tsv")

    data["QUERY"] = data["QUERY"].format(calibration_level)
    data["QUERY"] += (" AND  "
                      " CONTAINS( BOX('ICRS', {}, {}, {}, {}), "
                      " Plane.position_bounds ) = 1 ").format(ra.to(units.degree).value, dec.to(units.degree).value,
                                                              dra.to(units.degree).value, ddec.to(units.degree).value)
    if mjdate is not None:
        data["QUERY"] += " AND Plane.time_bounds_cval1 < {} AND Plane.time_bounds_cval2 > {} ".format(
            mjdate + 1.0 / 24.0,
            mjdate - 1 / 24.0)

    result = requests.get(TAP_WEB_SERVICE, params=data, verify=False)
    assert isinstance(result, requests.Response)
    logger.debug("Doing TAP Query using url: %s" % (str(result.url)))

    table_reader = ascii.get_reader(Reader=ascii.Basic)
    table_reader.header.splitter.delimiter = '\t'
    table_reader.data.splitter.delimiter = '\t'
    table = table_reader.read(result.text)

    logger.debug(str(table))
    return table


def populate(dataset_name, data_web_service_url=DATA_WEB_SERVICE + "CFHT"):
    """Given a dataset_name created the desired dbimages directories
    and links to the raw data files stored at CADC.
    @param dataset_name: the name of the CFHT dataset to make a link to.
    @param data_web_service_url: the URL of the data web service run by CADC.
    """

    data_dest = get_uri(dataset_name, version='o', ext='fits.fz')
    data_source = "%s/%so.fits.fz" % (data_web_service_url, dataset_name)

    mkdir(os.path.dirname(data_dest))

    try:
        vospace.client.link(data_source, data_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    header_dest = get_uri(dataset_name, version='o', ext='head')
    header_source = "%s/%so.fits.fz?cutout=[0]" % (
        data_web_service_url, dataset_name)
    try:
        vospace.client.link(header_source, header_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    header_dest = get_uri(dataset_name, version='p', ext='head')
    header_source = "%s/%s/%sp.head" % (
        'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/data/pub', 'CFHTSG', dataset_name)
    try:
        vospace.client.link(header_source, header_dest)
    except IOError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise e

    return True


def get_cands_uri(field, ccd, version='p', ext='measure3.cands.astrom', prefix=None,
                  block=None):
    """
        return the nominal URI for a candidate file.

    @param field: the OSSOS field name
    @param ccd: which CCD are the candidates on
    @param version: either the 'p', or 's' (scrambled) candidates.
    @param ext: Perhaps we'll change this one day.
    @param prefix: if this is a 'fake' dataset then add 'fk'
    @param block: Which BLOCK of the field are we looking at? eg. 15BS+1+1
    @return:
    """

    if prefix is None:
        prefix = ""
    if len(prefix) > 0:
        prefix += "_"
    if len(field) > 0:
        field += "_"

    if ext is None:
        ext = ""
    if len(ext) > 0 and ext[0] != ".":
        ext = ".{}".format(ext)

    measure3_dir = MEASURE3
    if block is not None:
        measure3_dir + "/{}".format(block)
    mkdir(measure3_dir)
    return "{}/{}{}{}{}{}".format(measure3_dir, prefix, field, version, ccd, ext)


def get_uri(expnum, ccd=None,
            version='p', ext='fits',
            subdir=None, prefix=None):
    """
    Build the uri for an OSSOS image stored in the dbimages
    containerNode.

    :rtype : basestring
    expnum: CFHT exposure number
    ccd: CCD in the mosaic [0-35]
    version: one of p,s,o etc.
    dbimages: dbimages containerNode.
    @type subdir: str
    @param expnum: int
    @param ccd:
    @param version:
    @param ext:
    @param subdir:
    @param prefix:
    """
    if subdir is None:
        subdir = str(expnum)
    if prefix is None:
        prefix = ''
    uri = os.path.join(DBIMAGES, subdir)

    if ext is None:
        ext = ''
    elif len(ext) > 0 and ext[0] != '.':
        ext = '.' + ext

    if version is None:
        version = ''

    if ccd is None:
        uri = os.path.join(uri,
                           '%s%s%s%s' % (prefix, str(expnum),
                                         version,
                                         ext))
    else:
        ccd = str(ccd).zfill(2)
        uri = os.path.join(uri,
                           'ccd{}'.format(ccd),
                           '%s%s%s%s%s' % (prefix, str(expnum),
                                           version,
                                           ccd,
                                           ext))

    return uri


dbimages_uri = get_uri


def set_tags_on_uri(uri, keys, values=None):
    node = vospace.client.get_node(uri)
    if values is None:
        values = []
        for idx in range(len(keys)):
            values.append(None)
    assert (len(values) == len(keys))
    for idx in range(len(keys)):
        key = keys[idx]
        value = values[idx]
        tag = tag_uri(key)
        node.props[tag] = value
    return vospace.client.add_props(node)


def _set_tags(expnum, keys, values=None):
    uri = os.path.join(DBIMAGES, str(expnum))
    node = vospace.client.get_node(uri, force=True)
    if values is None:
        values = []
        for idx in range(len(keys)):
            values.append(None)
    assert (len(values) == len(keys))
    for idx in range(len(keys)):
        key = keys[idx]
        tag = tag_uri(key)
        value = values[idx]
        node.props[tag] = value
    return vospace.client.add_props(node)


def set_tags(expnum, props):
    """Assign the key/value pairs in props as tags on on the given expnum.

    @param expnum: str
    @param props: dict
    @return: success
    """
    # now set all the props
    return _set_tags(expnum, props.keys(), props.values())


def set_tag(expnum, key, value):
    """Assign a key/value pair tag to the given expnum containerNode.

    @param expnum:  str
    @param key: str
    @param value: str
    @return: success
    """

    return set_tags(expnum, {key: value})


def tag_uri(key):
    """Build the uri for a given tag key. 

    @param key: what is the key that we need a stadanrd uri tag for? eg 'mkpsf_00'
    @type key: str
    """
    if OSSOS_TAG_URI_BASE in key:
        return key
    return OSSOS_TAG_URI_BASE + "#" + key.strip()


def get_tag(expnum, key):
    """given a key, return the vospace tag value.

    @param expnum: Number of the CFHT exposure that a tag value is needed for
    @param key: The process tag (such as mkpsf_00) that is being looked up.
    @return: the value of the tag
    @rtype: str
    """

    uri = tag_uri(key)
    force = uri not in get_tags(expnum)
    value = get_tags(expnum, force=force).get(uri, None)
    return value


def get_process_tag(program, ccd, version='p'):
    """
    make a process tag have a suffix indicating which ccd its for.
    @param program: Name of the process that a tag is built for.
    @param ccd: the CCD number that this process ran on.
    @param version: The version of the exposure (s, p, o) that the process ran on.
    @return: The string that represents the processing tag.
    """
    return "%s_%s%s" % (program, str(version), str(ccd).zfill(2))


def get_tags(expnum, force=False):
    """

    @param expnum:
    @param force:
    @return: dict
    @rtype: dict
    """
    uri = os.path.join(DBIMAGES, str(expnum))
    return vospace.client.get_node(uri, force=force).props


class Task(object):
    """
    A task within the OSSOS pipeline work-flow.
    """

    def __init__(self, executable, dependency=None):
        self.executable = executable
        self.name = os.path.splitext(self.executable)[0]
        self._target = None
        self._status = None
        self._dependency = None
        self.dependency = dependency

    def __str__(self):
        return self.name

    @property
    def tag(self):
        """
        Get the string representation of the tag used to annotate the status in VOSpace.
        @return: str
        """
        return "{}{}_{}{}{:02d}".format(self.target.prefix,
                                        self,
                                        self.target.version,
                                        self.target.ccd)

    @property
    def target(self):
        """

        @return: The target that this task is set to run on.
        @rtype: Target
        """
        return self._target

    @target.setter
    def target(self, target):
        assert isinstance(Target, target)
        self._target = target

    @property
    def dependency(self):
        """
        @rtype: Task
        """
        raise NotImplementedError()

    @dependency.setter
    def dependency(self, dependency):
        if dependency is None:
            self._dependency = dependency
        else:
            assert isinstance(Task, dependency)

    @property
    def status(self):
        """

        @return: The status of running this task on the given target.
        @rtype: str
        """
        return get_tag(self.target.expnum, self.tag)

    @status.setter
    def status(self, status):
        status += Time.now().iso
        set_tag(self.target.expnum, self.tag, status)

    @property
    def finished(self):
        """
        @rtype: bool
        """
        return self.status.startswith(SUCCESS)

    @property
    def ready(self):
        if self.dependency is None:
            return True
        else:
            return self.dependency.finished


class Target(object):
    """
    The target that a task will act on.
    """

    def __init__(self, prefix, expnum, version, ccd):
        """

        @param prefix: that is prefixed to the base exposure
        @type prefix: str
        @param expnum: the number of the CFHT exposure that is the target
        @type expnum: str
        @param version: Which version of the exposure (o, p, s) is the target.
        @type version: str
        @param ccd: which CCD of the exposure is the target.
        @type ccd: int
        """

        self.prefix = prefix
        self.expnum = expnum
        self.version = version
        self.ccd = ccd

    @property
    def name(self):
        return "{}{}{}{:02d}".format(self.prefix, self.expnum, self.version, self.ccd)

    @property
    def tags(self):
        """

        @rtype: list [str]
        """
        return get_tags(self.expnum)


def get_status(task, prefix, expnum, version, ccd, return_message=False):
    """
    Report back status of the given program by looking up the associated VOSpace annotation.

    @param task:  name of the process or task that will be checked.
    @param prefix: prefix of the file that was processed (often fk or None)
    @param expnum: which exposure number (or base filename)
    @param version: which version of that exposure (p, s, o)
    @param ccd: which CCD within the exposure.
    @param return_message: Return what did the TAG said or just /True/False/ for Success/Failure?
    @return: the status of the processing based on the annotation value.
    """
    key = get_process_tag(prefix+task, ccd, version)
    status = get_tag(expnum, key)
    logger.debug('%s: %s' % (key, status))
    if return_message:
        return status
    else:
        return status == SUCCESS


def set_status(task, prefix, expnum, version, ccd, status):
    """
    set the processing status of the given program.
    @param task: name of the processing task
    @param prefix: was there a prefix on the exposure number processed?
    @param expnum: exposure number processed.
    @param version: which version of the exposure? (p, s, o)
    @param ccd: the number of the CCD processing.
    @param status: What status to record:  "SUCCESS" we hope.
    @return: Success?
    """
    return set_tag(expnum, get_process_tag(prefix+task, ccd, version), status)


def get_file(expnum, ccd=None, version='p', ext='fits', subdir=None, prefix=None):
    uri = get_uri(expnum=expnum, ccd=ccd, version=version, ext=ext, subdir=subdir, prefix=prefix)
    filename = os.path.basename(uri)

    if not os.access(filename, os.F_OK):
        copy(uri, filename)

    return filename


def decompose_content_decomposition(content_decomposition):
    """

    :param content_decomposition:
    :return:
    """
    # check for '-*' in the cutout string and replace is naxis:1
    content_decomposition = re.findall('(\d+)__(\d*)_(\d*)_(\d*)_(\d*)', content_decomposition)
    if len(content_decomposition) == 0:
        content_decomposition = [(0, 1, -1, 1, -1)]
    return content_decomposition


def _cutout_expnum(observation, sky_coord, radius):
    """
    Get a cutout from an exposure based on the RA/DEC location.
    @param observation: The Observation object that contains the expusre number information.
    @type observation: Observation
    @param sky_coord: which RA/DEC is needed,
    @type sky_coord:  SkyCoord
    @param radius:
    @type radius: Quantity
    @return: HDUList containing the cutout image.
    @rtype: list(HDUList)
    """

    uri = observation.get_image_uri()
    base_url = ("http://{}/vospace/auth/nodes/{}?view=cutout&"
                "cutout=CIRCLE+ICRS+{}+{}+{}")
    url = base_url.format(vospace.VOSPACE_SERVER,
                          uri[4:],
                          sky_coord.ra.to('degree').value,
                          sky_coord.dec.to('degree').value,
                          radius.to('degree').value)
    resp = requests.get(url, auth=vospace.authentication)
    hdulist = fits.open(cStringIO.StringIO(resp.content))

    hdulist.verify('silentfix+ignore')
    disposition = resp.headers.get('content-disposition', '0___')
    cutouts = decompose_content_decomposition(disposition)
    logger.debug("Got cutout boundaries of: {}".format(cutouts))
    logger.debug("Initial Length of HDUList: {}".format(len(hdulist)))

    # Make sure here is a primaryHDU
    if len(hdulist) == 1:
        phdu = fits.PrimaryHDU()
        phdu.header['ORIGIN'] = "OSSOS"
        hdulist.insert(0, phdu)

    logger.debug("Final Length of HDUList: {}".format(len(hdulist)))

    if len(cutouts) != len(hdulist) - 1:
        raise ValueError("Wrong number of cutout structures found in Content-Disposition response.")

    for hdu in hdulist[1:]:
        cutout = cutouts.pop(0)
        if 'ASTLEVEL' not in hdu.header:
            print "WARNING: ******* NO ASTLEVEL KEYWORD ********** for {0} ********".format(observation.get_image_uri)
            hdu.header['ASTLEVEL'] = 0
        hdu.header['EXTNO'] = cutout[0]
        naxis1 = hdu.header['NAXIS1']
        naxis2 = hdu.header['NAXIS2']
        default_datasec = "[{}:{},{}:{}]".format(1, naxis1, 1, naxis2)
        datasec = hdu.header.get('DATASEC', default_datasec)
        datasec = datasec_to_list(datasec)
        corners = datasec
        for idx in range(len(corners)):
            try:
                corners[idx] = int(cutout[idx+1])
            except Exception as ex:
                pass

        hdu.header['DATASEC'] = reset_datasec("[{}:{},{}:{}]".format(corners[0],
                                                                     corners[1],
                                                                     corners[2],
                                                                     corners[3]),
                                              hdu.header.get('DATASEC', default_datasec),
                                              hdu.header['NAXIS1'],
                                              hdu.header['NAXIS2'])
        hdu.header['XOFFSET'] = int(corners[0]) - 1
        hdu.header['YOFFSET'] = int(corners[2]) - 1
        hdu.converter = CoordinateConverter(hdu.header['XOFFSET'], hdu.header['YOFFSET'])
        try:
                hdu.wcs = WCS(hdu.header)
        except Exception as ex:
            logger.error("Failed trying to initialize the WCS for {}".format(uri))
            raise ex
    logger.debug("Sending back {}".format(hdulist))
    return hdulist


def ra_dec_cutout(uri, sky_coord, radius):
    """

    :param uri: The vospace location of the image to make a cutout from
    :type uri: str
    :param sky_coord: The central coordinate of the cutout
    :type sky_coord: SkyCoord
    :param radius: The radius of the cutout
    :type radius: Quantity
    :return: An HDUList with the cutout
    :rtype: fits.HDUList
    """

    # These two lines enable debugging at httplib level (requests->urllib3->http.client)
    # You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
    # The only thing missing will be the response.body which is not logged.
    try:
        import http.client as http_client
    except ImportError:
        # Python 2
        import httplib as http_client

    # Get the 'uncut' images CRPIX1/CRPIX2 values

    coo_sys = upper(sky_coord.frame.name)
    this_cutout = "CIRCLE {} {} {} {}".format(coo_sys, sky_coord.ra.to(units.degree).value,
                                              sky_coord.dec.to(units.degree).value,
                                              radius.to(units.degree).value)

    view = "cutout"
    url = "http://{}/vospace/auth/nodes/{}".format(vospace.VOSPACE_SERVER,
                                                   uri[4:])
    params = {"view": view,
              "cutout": this_cutout}
    resp = requests.get(url, params, auth=vospace.authentication)
    fobj = cStringIO.StringIO(resp.content)
    fobj.seek(0)
    hdulist = fits.open(fobj)
    hdulist.verify('silentfix+ignore')
    cutouts = decompose_content_decomposition(resp.headers.get('content-disposition', '0___'))
    logger.debug("Got cutout boundaries of: {}".format(cutouts))
    logger.debug("Initial Length of HDUList: {}".format(len(hdulist)))

    # Make sure here is a primaryHDU
    if len(hdulist) == 1:
        phdu = fits.PrimaryHDU()
        phdu.header['ORIGIN'] = "OSSOS"
        hdulist.insert(0, phdu)

    logger.debug("Final Length of HDUList: {}".format(len(hdulist)))

    if len(cutouts) != len(hdulist) - 1:
        raise ValueError("Wrong number of cutout structures found in Content-Disposition response.")

    for hdu in hdulist[1:]:
        cutout = cutouts.pop(0)
        if 'ASTLEVEL' not in hdu.header:
            print "******* NO ASTLEVEL ****************** for {0} ********".format(uri)
            hdu.header['ASTLEVEL'] = 0
        hdu.header['EXTNO'] = cutout[0]
        hdu.header['DATASEC'] = reset_datasec("[{}:{},{}:{}]".format(cutout[1],
                                                                     cutout[2],
                                                                     cutout[3],
                                                                     cutout[4]),
                                              hdu.header.get('DATASEC', None),
                                              hdu.header['NAXIS1'],
                                              hdu.header['NAXIS2'])
        hdu.header['XOFFSET'] = int(cutout[1]) - 1
        hdu.header['YOFFSET'] = int(cutout[3]) - 1

        hdu.converter = CoordinateConverter(hdu.header['XOFFSET'], hdu.header['YOFFSET'])
        try:
                hdu.wcs = WCS(hdu.header)
        except Exception as ex:
            logger.error("Failed trying to initialize the WCS for {}".format(uri))
            raise ex
    logger.debug("Sending back {}".format(hdulist))
    return hdulist


def get_image(expnum, ccd=None, version='p', ext='fits',
              subdir=None, prefix=None, cutout=None, return_file=True, flip_image=True):
    """Get a FITS file for this expnum/ccd  from VOSpace.


    @param cutout: (str)
    @param return_file: return an filename (True) or HDUList (False)
    @param expnum: CFHT exposure number (int)
    @param ccd:     @param ccd:
    @param version: [p, s, o]  (char)
    @param ext:
    @param subdir:
    @param prefix:
    @param flip_image: Should the image be x/y flipped after being retrieved?
    @return: astropy.io.fits.PrimaryHDU
    """

    filename = os.path.basename(get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix))
    if os.access(filename, os.F_OK) and return_file and cutout is None:
        return filename

    cutout_string = cutout
    try:
        if os.access(filename, os.F_OK) and cutout:
            cutout = datasec_to_list(cutout)
            hdulist = fits.open(filename)
            if len(hdulist) > 1:
                raise ValueError("Local cutout access not designed to work on MEFs yet.")
            header = hdulist[0].header
            cutout[0] = cutout[0] < 0 and header['NAXIS1'] - cutout[0] + 1 or cutout[0]
            cutout[1] = cutout[1] < 0 and header['NAXIS1'] - cutout[1] + 1 or cutout[1]
            cutout[2] = cutout[2] < 0 and header['NAXIS2'] - cutout[2] + 1 or cutout[2]
            cutout[3] = cutout[3] < 0 and header['NAXIS2'] - cutout[3] + 1 or cutout[3]
            logger.debug("DATA array shape: {}".format(hdulist[0].data.shape))
            logger.debug("CUTOUT array: {} {} {} {}".format(cutout[0],
                                                            cutout[1],
                                                            cutout[2],
                                                            cutout[3]))
            flip = cutout[0] < cutout[1] and 1 or -1
            flop = cutout[2] < cutout[3] and 1 or -1
            header['CRPIX1'] = (header.get("CRPIX1", 1) - cutout[0]) * flip + 1
            header['CRPIX2'] = (header.get("CRPIX2", 1) - cutout[2]) * flop + 1
            header['CD1_1'] = header.get("CD1_1", 1) * flip
            header['CD2_1'] = header.get("CD2_1", 1) * flip
            header['CD2_2'] = header.get("CD2_2", 1) * flop
            header['CD1_2'] = header.get("CD1_2", 1) * flop

            data = hdulist[0].data[cutout[2]-1:cutout[3], cutout[0]-1:cutout[1]]
            hdulist[0].data = data
            header['DATASEC'] = reset_datasec(cutout_string,
                                              header['DATASEC'],
                                              header['NAXIS1'],
                                              header['NAXIS2'])
            if return_file:
                cutout_filename = os.path.splitext(filename)[0]+"_{}_{}_{}_{}.fits".format(cutout[0],
                                                                                           cutout[1],
                                                                                           cutout[2],
                                                                                           cutout[3])
                hdulist.writeto(cutout_filename)
                return cutout_filename
            else:
                return hdulist
    except Exception as e:
        logger.debug(str(e))
        logger.debug("Failed trying to access local copy: {} with cutout [{}:{}, {}:{}], using VOSpace".format(
            filename, cutout[2]-1, cutout[3], cutout[0]-1, cutout[1]))

    cutout = cutout_string

    if not subdir:
        subdir = str(expnum)

    logger.debug("Building list of possible uri locations")
    # # here is the list of places we will look, in order
    if version != 'p':
        locations = [(get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix),
                      cutout)]
    else:
        locations = []
    logger.debug(str(locations))
    if ccd is not None:
        try:
            for this_ext in [ext, ext + ".fz"]:
                ext_no = int(ccd) + 1
                # extension 1 -> 18 +  37,36 should be flipped.
                flip_these_extensions = range(1, 19)
                flip_these_extensions.append(37)
                flip_these_extensions.append(38)
                flip = (cutout is None and "fits" in ext and (
                    (ext_no in flip_these_extensions and flip_image) and "[-*,-*]" or "[*,*]")) or cutout
                locations.append((get_uri(expnum, version=version, ext=this_ext, subdir=subdir),
                                  "[{}]{}".format(ext_no, flip)))
        except Exception as e:
            logger.error(str(e))
            pass
    else:
        uri = get_uri(expnum, ccd, version, ext=ext, subdir=subdir, prefix=prefix)
        locations.append((uri, cutout))
        uri = get_uri(expnum, ccd, version, ext=ext + ".fz", subdir=subdir, prefix=prefix)
        locations.append((uri, cutout))
    while len(locations) > 0:
        (uri, cutout) = locations.pop(0)
        try:
            hdu_list = get_hdu(uri, cutout)
            if return_file:
                hdu_list.writeto(filename)
                del hdu_list
                return filename
            else:
                return hdu_list
        except Exception as e:
            logging.warning("Failed to get image using: {}{}".format(uri, cutout))
            logger.debug("{}".format(type(e)))
            logger.debug("Failed to open {} cutout:{}".format(uri, cutout))
            logger.debug("vos sent back error: {} code: {}".format(str(e), getattr(e, 'errno', 0)))
    raise IOError(errno.ENOENT, "Failed to get image using {} {} {} {}.".format(expnum, version, ccd, cutout))


def datasec_to_list(datasec):
    """
    convert an IRAF style PIXEL DATA section as to a list of integers.
    @param datasec: str
    @return: list
    """

    return [int(x) for x in re.findall(r"([-+]?[\*\d]+?)[:,\]]+", datasec)]


def reset_datasec(cutout, datasec, naxis1, naxis2):
    """
    reset the datasec to account for a possible cutout.

    @param cutout:
    @param datasec:
    @param naxis1: size of the original image in the 'x' direction
    @param naxis2: size of the oringal image in the 'y' direction
    @return:
    """

    if cutout is None or cutout == "[*,*]":
        return datasec
    try:
        datasec = datasec_to_list(datasec)
    except:
        return datasec

    # check for '-*' in the cutout string and replace is naxis:1
    cutout = cutout.replace(" ", "")
    cutout = cutout.replace("[-*,", "{}:1,".format(naxis1))
    cutout = cutout.replace(",-*]", ",{}:1]".format(naxis2))
    cutout = cutout.replace("[*,", "[1:{},".format(naxis1))
    cutout = cutout.replace(",*]", ",1:{}]".format(naxis1))

    try:
        cutout = [int(x) for x in re.findall(r"([-+]?[\*\d]+?)[:,\]]+", cutout)]
    except:
        logger.debug("Failed to processes the cutout pattern: {}".format(cutout))
        return datasec

    if len(cutout) == 5:
        # cutout likely starts with extension, remove
        cutout = cutout[1:]

    # -ve integer offsets indicate offset from the end of array.
    for idx in [0, 1]:
        if cutout[idx] < 0:
            cutout[idx] = naxis1 - cutout[idx] + 1
    for idx in [2, 3]:
        if cutout[idx] < 0:
            cutout[idx] = naxis2 - cutout[idx] + 1

    flip = cutout[0] > cutout[1]
    flop = cutout[2] > cutout[3]

    logger.debug("Working with cutout: {}".format(cutout))

    if flip:
        cutout = [naxis1 - cutout[0] + 1, naxis1 - cutout[1] + 1, cutout[2], cutout[3]]
        datasec = [naxis1 - datasec[1] + 1, naxis1 - datasec[0] + 1, datasec[2], datasec[3]]

    if flop:
        cutout = [cutout[0], cutout[1], naxis2 - cutout[2] + 1, naxis2 - cutout[3] + 1]
        datasec = [datasec[0], datasec[1], naxis2 - datasec[3] + 1, naxis2 - datasec[2] + 1]

    datasec = [max(datasec[0] - cutout[0] + 1, 1),
               min(datasec[1] - cutout[0] + 1, naxis1),
               max(datasec[2] - cutout[2] + 1, 1),
               min(datasec[3] - cutout[2] + 1, naxis2)]

    return "[{}:{},{}:{}]".format(datasec[0], datasec[1], datasec[2], datasec[3])


def get_hdu(uri, cutout=None):
    """Get a at the given uri from VOSpace, possibly doing a cutout.

    If the cutout is flips the image then we also must flip the datasec keywords.  Also, we must offset the
    datasec to reflect the cutout area being used.

    @param uri: The URI in VOSpace of the image to HDU to retrieve.
    @param cutout: A CADC data service CUTOUT paramter to be used when retrieving the observation.
    @return: fits.HDU
    """
    try:
        # the filename is based on the Simple FITS images file.
        filename = os.path.basename(uri)
        if os.access(filename, os.F_OK) and cutout is None:
            logger.debug("File already on disk: {}".format(filename))
            hdu_list = fits.open(filename, scale_back=True)
            hdu_list.verify('silentfix+ignore')

        else:
            logger.debug("Pulling: {}{} from VOSpace".format(uri, cutout))
            fpt = tempfile.NamedTemporaryFile(suffix='.fits')
            cutout = cutout is not None and cutout or ""
            vospace.client.copy(uri+cutout, fpt.name)
            fpt.seek(0, 2)
            fpt.seek(0)
            logger.debug("Read from vospace completed. Building fits object.")
            hdu_list = fits.open(fpt, scale_back=False)
            hdu_list.verify('silentfix+ignore')

            logger.debug("Got image from vospace")
            try:
                hdu_list[0].header['DATASEC'] = reset_datasec(cutout, hdu_list[0].header['DATASEC'],
                                                              hdu_list[0].header['NAXIS1'],
                                                              hdu_list[0].header['NAXIS2'])
            except Exception as e:
                logging.debug("error converting datasec: {}".format(str(e)))

        for hdu in hdu_list:
            logging.debug("Adding converter to {}".format(hdu))
            hdu.converter = CoordinateConverter(0, 0)
            try:
                hdu.wcs = WCS(hdu.header)
            except Exception as ex:
                logger.error("Failed trying to initialize the WCS: {}".format(ex))
    except Exception as ex:
        print ex
        raise ex
    return hdu_list


def get_trans(expnum, ccd, prefix=None, version='p'):
    """

    @param expnum: The exposure number to get the WCS transformation for.
    @param ccd: The CCD of that exposure
    @param prefix: Any possible prefixs (such as 'fk')
    @param version: which version of the exposure?
    @return:
    """
    uri = get_uri(expnum, ccd, version, ext='trans.jmp', prefix=prefix)
    logging.info("get_trans: {}".format(uri))
    fobj = open_vos_or_local(uri)
    line = fobj.read()
    fobj.close()
    vs = line.split()
    trans = {'dx': float(vs[0]),
             'cd11': float(vs[1]),
             'cd12': float(vs[2]),
             'dy': float(vs[3]),
             'cd21': float(vs[4]),
             'cd22': float(vs[5])}
    return trans

def get_fwhm_tag(expnum, ccd, prefix=None, version='p'):
    """
    Get the FWHM from the VOSpace annotation.

    @param expnum:
    @param ccd:
    @param prefix:
    @param version:
    @return:
    """
    uri = get_uri(expnum, ccd, version, ext='fwhm', prefix=prefix)
    if uri not in fwhm:
        key = "fwhm_{:1s}{:02d}".format(version, int(ccd))
        fwhm[uri] = get_tag(expnum, key)
    return fwhm[uri]

def get_fwhm(expnum, ccd, prefix=None, version='p'):
    """Get the FWHM computed for the given expnum/ccd combo.

    @param expnum:
    @param ccd:
    @param prefix:
    @param version:
    @return:
    """

    uri = get_uri(expnum, ccd, version, ext='fwhm', prefix=prefix)
    filename = os.path.basename(uri)

    try:
        return fwhm[uri]
    except:
        pass

    try:
        fwhm[uri] = float(open(filename, 'r').read())
        return fwhm[uri]
    except:
        pass

    try:
        fwhm[uri] = float(open_vos_or_local(uri).read())
        return fwhm[uri]
    except Exception as ex:
        logger.error(str(ex))
        fwhm[uri] = 4.0
        return fwhm[uri]


def _get_zeropoint(expnum, ccd, prefix=None, version='p'):
    """
    Retrieve the zeropoint stored in the tags associated with this image.
    @param expnum: Exposure number
    @param ccd: ccd of the exposure
    @param prefix: possible prefix (such as 'fk')
    @param version: which version: p, s, or o ?
    @return: zeropoint
    """
    if prefix is not None:
        DeprecationWarning("Prefix is no longer used here as the 'fk' and 's' have the same zeropoint.")

    key = "zeropoint_{:1s}{:02d}".format(version, int(ccd))
    return get_tag(expnum, key)


def get_zeropoint(expnum, ccd, prefix=None, version='p'):
    """Get the zeropoint for this exposure using the zeropoint.used file created during source planting..

    This command expects that there is a file called #######p##.zeropoint.used which contains the zeropoint.

    @param expnum: exposure to get zeropoint of
    @param ccd: which ccd (extension - 1) to get zp
    @param prefix: possible string prefixed to expsoure number.
    @param version:  one of [spo] as in #######p##
    @return: zeropoint
    """
    uri = get_uri(expnum, ccd, version, ext='zeropoint.used', prefix=prefix)
    try:
        return zmag[uri]
    except:
        pass

    try:
        zmag[uri] = float(open_vos_or_local(uri).read())
        return zmag[uri]
    except:
        pass

    zmag[uri] = 0.0
    return zmag[uri]


def mkdir(dirname):
    """make directory tree in vospace.

    @param dirname: name of the directory to make
    """
    dir_list = []

    while not vospace.client.isdir(dirname):
        dir_list.append(dirname)
        dirname = os.path.dirname(dirname)
    while len(dir_list) > 0:
        logging.info("Creating directory: %s" % (dir_list[-1]))
        try:
            vospace.client.mkdir(dir_list.pop())
        except IOError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise e


def vofile(filename, **kwargs):
    """
    Open and return a handle on a VOSpace data connection
    @param filename:
    @param kwargs:
    @return:
    """
    basename = os.path.basename(filename)
    if os.access(basename, os.R_OK):
        return open(basename, 'r')
    kwargs['view'] = kwargs.get('view', 'data')
    try:
        url = "http://{}/data/auth/vospace/{}".format(vospace.VOSPACE_SERVER,
                                                      filename[4:])
        return cStringIO.StringIO(requests.get(url,
                                               params=kwargs,
                                               auth=vospace.authentication).content)
    except Exception as ex:
        logging.info("Error using quick vofile access: {}".format(ex))
        return vospace.client.open(filename, **kwargs)


def open_vos_or_local(path, mode="rb"):
    """
    Opens a file which can either be in VOSpace or the local filesystem.

    @param path:
    @param mode:
    @return:
    """
    if path.startswith("vos:"):
        primary_mode = mode[0]
        if primary_mode == "r":
            vofile_mode = os.O_RDONLY
        elif primary_mode == "w":
            vofile_mode = os.O_WRONLY
        elif primary_mode == "a":
            vofile_mode = os.O_APPEND
        else:
            raise ValueError("Can't open with mode %s" % mode)

        return vofile(path, mode=vofile_mode)
    else:
        return open(path, mode)


def copy(source, dest):
    """
    use the vospace service to get a file.

    @param source:
    @param dest:
    @return:
    """
    logger.info("copying {} -> {}".format(source, dest))

    return vospace.client.copy(source, dest)


def vlink(s_expnum, s_ccd, s_version, s_ext,
          l_expnum, l_ccd, l_version, l_ext, s_prefix=None, l_prefix=None):
    """make a link between two version of a file.

    @param s_expnum:
    @param s_ccd:
    @param s_version:
    @param s_ext:
    @param l_expnum:
    @param l_ccd:
    @param l_version:
    @param l_ext:
    @param s_prefix:
    @param l_prefix:
    @return:
    """
    source_uri = get_uri(s_expnum, ccd=s_ccd, version=s_version, ext=s_ext, prefix=s_prefix)
    link_uri = get_uri(l_expnum, ccd=l_ccd, version=l_version, ext=l_ext, prefix=l_prefix)

    return vospace.client.link(source_uri, link_uri)


def remove(uri):
    try:
        vospace.client.delete(uri)
    except Exception as e:
        logger.debug(str(e))


def delete(expnum, ccd, version, ext, prefix=None):
    """
    delete a file, no error on does not exist

    @param expnum:
    @param ccd:
    @param version:
    @param ext:
    @param prefix:
    @return:
    """
    uri = get_uri(expnum, ccd=ccd, version=version, ext=ext, prefix=prefix)
    remove(uri)


def my_glob(pattern):
    """
    get a listing matching pattern

    @param pattern:
    @return:
    """
    result = []
    if pattern[0:4] == 'vos:':
        dirname = os.path.dirname(pattern)
        flist = listdir(dirname)
        for fname in flist:
            fname = '/'.join([dirname, fname])
            if fnmatch.fnmatch(fname, pattern):
                result.append(fname)
    else:
        result = glob(pattern)
    return result


def listdir(directory, force=False):
    return vospace.client.listdir(directory, force=force)


def list_dbimages():
    return listdir(DBIMAGES)


def exists(uri, force=False):
    try:
        return vospace.client.get_node(uri, force=force) is not None
    except EnvironmentError as e:
        logger.error(str(e))  # not critical enough to raise
        # Sometimes the error code returned is the OS version, sometimes the HTTP version
        if e.errno in [404, os.errno.ENOENT]:
            return False


def move(old_uri, new_uri):
    vospace.client.move(old_uri, new_uri)


def delete_uri(uri):
    vospace.client.delete(uri)


def has_property(node_uri, property_name, ossos_base=True):
    """
    Checks if a node in VOSpace has the specified property.

    @param node_uri:
    @param property_name:
    @param ossos_base:
    @return:
    """
    if get_property(node_uri, property_name, ossos_base) is None:
        return False
    else:
        return True


def get_property(node_uri, property_name, ossos_base=True):
    """
    Retrieves the value associated with a property on a node in VOSpace.

    @param node_uri:
    @param property_name:
    @param ossos_base:
    @return:
    """
    # Must use force or we could have a cached copy of the node from before
    # properties of interest were set/updated.
    node = vospace.client.get_node(node_uri, force=True)
    property_uri = tag_uri(property_name) if ossos_base else property_name

    if property_uri not in node.props:
        return None

    return node.props[property_uri]


def set_property(node_uri, property_name, property_value, ossos_base=True):
    """
    Sets the value of a property on a node in VOSpace.  If the property
    already has a value then it is first cleared and then set.

    @param node_uri:
    @param property_name:
    @param property_value:
    @param ossos_base:
    @return:
    """
    node = vospace.client.get_node(node_uri)
    property_uri = tag_uri(property_name) if ossos_base else property_name

    # If there is an existing value, clear it first
    if property_uri in node.props:
        node.props[property_uri] = None
        vospace.client.add_props(node)

    node.props[property_uri] = property_value
    vospace.client.add_props(node)


def build_counter_tag(epoch_field, dry_run=False):
    """
    Builds the tag for the counter of a given epoch/field, without the OSSOS base.

    @param epoch_field:
    @param dry_run:
    @return:
    """
    logger.info("Epoch Field: {}, OBJECT_COUNT {}".format(str(epoch_field), str(OBJECT_COUNT)))
    tag = epoch_field[1] + "-" + OBJECT_COUNT

    if dry_run:
        tag += "-DRYRUN"

    return tag


def read_object_counter(node_uri, epoch_field, dry_run=False):
    """
    Reads the object counter for the given epoch/field on the specified node.

    @param node_uri:
    @param epoch_field:
    @param dry_run:
    @return: the current object count.
    """
    return get_property(node_uri, build_counter_tag(epoch_field, dry_run),
                        ossos_base=True)


def increment_object_counter(node_uri, epoch_field, dry_run=False):
    """
    Increment the object counter used to create unique object identifiers.

    @param node_uri:
    @param epoch_field:
    @param dry_run:
    @return: The object count AFTER incrementing.

    """
    current_count = read_object_counter(node_uri, epoch_field, dry_run=dry_run)

    if current_count is None:
        new_count = "01"
    else:
        new_count = coding.base36encode(coding.base36decode(current_count) + 1,
                                        pad_length=2)

    set_property(node_uri,
                 build_counter_tag(epoch_field, dry_run=dry_run),
                 new_count,
                 ossos_base=True)

    return new_count


def get_mopheader(expnum, ccd, version='p', prefix=None):
    """
    Retrieve the mopheader, either from cache or from vospace

    @param expnum:
    @param ccd:
    @param version:
    @param prefix:
    @return: Header
    """
    prefix = prefix is None and "" or prefix
    mopheader_uri = dbimages_uri(expnum=expnum,
                                 ccd=ccd,
                                 version=version,
                                 prefix=prefix,
                                 ext='.mopheader')
    if mopheader_uri in mopheaders:
        return mopheaders[mopheader_uri]

    filename = os.path.basename(mopheader_uri)

    if os.access(filename, os.F_OK):
        logger.debug("File already on disk: {}".format(filename))
        mopheader_fpt = cStringIO.StringIO(open(filename, 'r').read())
    else:
        mopheader_fpt = cStringIO.StringIO(open_vos_or_local(mopheader_uri).read())

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', AstropyUserWarning)
        mopheader = fits.open(mopheader_fpt)

        # add some values to the mopheader so it can be an astrom header too.
        header = mopheader[0].header
        try:
            header['FWHM'] = get_fwhm(expnum, ccd)
        except IOError:
            header['FWHM'] = 10
        header['SCALE'] = mopheader[0].header['PIXSCALE']
        header['NAX1'] = header['NAXIS1']
        header['NAX2'] = header['NAXIS2']
        header['MOPversion'] = header['MOP_VER']
        header['MJD_OBS_CENTER'] = str(Time(header['MJD-OBSC'],
                                            format='mjd',
                                            scale='utc', precision=5).replicate(format='mpc'))
        header['MAXCOUNT'] = MAXCOUNT
        mopheaders[mopheader_uri] = header
        mopheader.close()
    return mopheaders[mopheader_uri]


def _get_sghead(expnum):
    """
    Use the data web service to retrieve the stephen's astrometric header.

    :param expnum: CFHT exposure number you want the header for
    :rtype : list of astropy.io.fits.Header objects.
    """
    version = 'p'

    key = "{}{}".format(expnum, version)
    if key in sgheaders:
        return sgheaders[key]

    url = "http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/data/pub/CFHTSG/{}{}.head".format(expnum, version)
    logging.getLogger("requests").setLevel(logging.ERROR)
    resp = requests.get(url)
    if resp.status_code != 200:
        raise IOError(errno.ENOENT, "Could not get {}".format(url))

    header_str_list = re.split('END      \n', resp.content)

    # # make the first entry in the list a Null
    headers = [None]
    for header_str in header_str_list:
        headers.append(fits.Header.fromstring(header_str, sep='\n'))
    sgheaders[key] = headers
    return sgheaders[key]


def get_header(uri):
    """
    Pull a FITS header from observation at the given URI

    @param uri:  The URI of the image in VOSpace.
    """
    if uri not in astheaders:
        astheaders[uri] = get_hdu(uri, cutout="[1:1,1:1]")[0].header
    return astheaders[uri]


def get_astheader(expnum, ccd, version='p', prefix=None):
    """
    Retrieve the header for a given dbimages file.

    @param expnum:  CFHT odometer number
    @param ccd: which ccd based extension (0..35)
    @param version: 'o','p', or 's'
    @param prefix:
    @return:
    """
    logger.debug("Getting ast header for {}".format(expnum))
    if version == 'p':
        # Get the SG header if possible.
        sg_key = "{}{}".format(expnum, version)
        if sg_key not in sgheaders:
            _get_sghead(expnum)
        if sg_key in sgheaders:
            for header in sgheaders[sg_key]:
                try:
                    if header.get('EXTVER', -1) == int(ccd):
                        return header
                except:
                    pass

    ast_uri = dbimages_uri(expnum, ccd, version=version, ext='.fits')
    if ast_uri not in astheaders:
        try:
            url = 'http://{}/data/auth/vospace/{}'.format(
                vospace.VOSPACE_SERVER,
                ast_uri[4:])
            resp = requests.get(url, params={'view': 'cutout', 'cutout': '[1:1,1:1]'},
                                auth=vospace.authentication)
            hdulist = fits.open(cStringIO.StringIO(resp.content))
        except Exception as ex:
            logging.info("Failed: {}".format(ex))
            hdulist = get_image(expnum, ccd=ccd, version=version, prefix=prefix,
                                cutout="[1:1,1:1]", return_file=False)
        assert isinstance(hdulist, fits.HDUList)
        astheaders[ast_uri] = hdulist[0].header
    return astheaders[ast_uri]


def log_filename(prefix, task, version, ccd):
    return "{}{}_{}{}.txt".format(prefix, task, version, ccd)


def log_location(expnum, ccd):
    return os.path.dirname(get_uri(expnum, ccd=ccd))


def set_logger(task, prefix, expnum, ccd, version, dry_run):
    this_logger = logging.getLogger()
    log_format = logging.Formatter('%(asctime)s - %(module)s.%(funcName)s %(lineno)d: %(message)s')

    filename = log_filename(prefix, task, ccd=ccd, version=version)
    location = log_location(expnum, ccd)
    if not dry_run:
        vo_handler = util.VOFileHandler("/".join([location, filename]))
        vo_handler.setFormatter(log_format)
        this_logger.addHandler(vo_handler)

    file_handler = logging.FileHandler(filename=filename)
    file_handler.setFormatter(log_format)
    this_logger.addHandler(file_handler)



import fcntl
import os
from numpy import random
from astropy.table import Table
import numpy
import storage
from scipy import interpolate


class MatchFile(object):

    def __init__(self, prefix, field, ccd, version, ext):
        self.field = field
        self.ccd = ccd
        self.version = version
        self.prefix = prefix
        self.ext = "measure3.{}.astrom".format(ext)
        self.match_filename = 'test.txt'

    @property
    def filename(self):
        return os.path.basename(storage.get_cands_uri(self.field,
                                                      ccd=self.ccd,
                                                      version=self.version,
                                                      prefix=self.prefix,
                                                      ext="measure3.{}.match".format(self.ccd)))

    def __enter__(self):
        """Acquire a lock on the output file, prevents collisions between multiple runs."""
        self.fd = open(self.filename, 'a')
        fcntl.lockf(self.fd, fcntl.LOCK_EX)
        return self.fd

    def __exit__(self):
        self.fd.close()


class Range(object):
    """A custom object that is initialized with a range and when called returns a random value in that range."""

    def __init__(self, minimum, maximum=None, seed=None, func=None):
        random.seed(seed)
        if maximum is None:
            if len(minimum) == 2:
                maximum = minimum[1]
                minimum = minimum[0]
            else:
                raise ValueError("Range should be initialize with a tuple or two arguments giving min and max.")

        self.min = minimum
        self.max = maximum
        self.value = None
        self._dist = None
        self.func = func

    @property
    def dist(self):
        """
        How should the values in range be sampled, uniformly or via some function.
        :return:
        """
        if self.func is None:
            return random.uniform(self.min, self.max)
        if self._dist is None:
            x = numpy.arange(self.min, self.max, (self.max-self.min)/1000.0)
            p = self.func(x).cumsum()
            p -= p.min()
            p /= p.max()
            self._dist = interpolate.interp1d(p, x)
        return self._dist(random.random())

    def __call__(self, new=True):
        if new or self.value is None:
            self.value = self.dist
        return self.value

    def __eq__(self, other):
        return float(self) == float(other)

    def __cmp__(self, other):
        return cmp(float(self), float(other))

    def __float__(self):
        return self.value

    def __str__(self):
        if self.value is None:
            self.__call__()
        return str(self.value)


class KBOGenerator(object):

    def __init__(self, n, rate, angle, mag, x, y, id0=None):

        self.rate = rate
        self.angle = angle
        self.mag = mag
        self.x = x
        self.y = y
        self._id = id0
        self._n = n

    def __iter__(self):
        return self

    @property
    def id(self):
        if self._id is None:
            self._id = 0
        self._id += 1
        return self._id

    def next(self):
        """
        :return: a set of values that can be used for an planted object builder.
        """
        #        x          y        mag   pix rate      angle  ''/h rate     id
        # 912.48     991.06      22.01      57.32     -45.23      10.60      0
        self._n -= 1
        if self._n < 0:
            raise StopIteration()
        return {'x': self.x(), 'y': self.y(), 'mag': self.mag(), 'sky_rate': self.rate(), 'angle': self.angle(),
                'id': self.id}

    @classmethod
    def _step(cls, mag):
        low = mag * 0 + 0.3
        high = mag * 0 + 0.7
        g = [mag < 23.3] * low + [mag >= 23.3] * high
        return g[0]

    @classmethod
    def get_kbos(cls, n, rate, angle, mag, x, y, filename=None):

        kbos = Table(names=('x', 'y', 'mag', 'sky_rate', 'angle', 'id'))

        # generate the KBOs.
        for kbo in cls(n,
                       rate=Range(rate, func=lambda value: value**0.25),
                       angle=Range(angle),
                       mag=Range(mag, func=cls._step),
                       x=Range(x),
                       y=Range(y)):
            kbos.add_row(kbo)

        # Write to a local file if filename given.
        if filename is not None:
            fd = open(filename, 'w')
            fd.write("#K {:10s} {:10s} {:10s} {:10s} {:10s}\n".format("rate", "angle", "mag", "x_range", "y_range"))
            fd.write("#V {:10s} {:10s} {:10s} {:10s} {:10s}\n".format(rate, angle, mag, x, y))
            fd.write("# ")
            kbos.write(fd, format='ascii.fixed_width', delimiter=None)
            fd.close()

        return kbos



"""This module abstracts all vospace activities.  Including a switch to using username/password pairs."""
from getpass import getpass

from requests.auth import HTTPBasicAuth
from vos.vos import Client, Connection
import sys
import types
import netrc
import logging
logging.getLogger('vos').setLevel(logging.ERROR)

VOSPACE_SERVER = "www.canfar.phys.uvic.ca"


class Wrapper(Client):

    def __getattribute__(self, item):
        func = object.__getattribute__(self, item)
        if isinstance(func, types.MethodType):
            def my_func(*args, **kwargs):
                sys.stderr.write("{} -> {} {}\n".format(item, args[1:], kwargs))
                result = func(*args[1:], **kwargs)
                sys.stderr.write("{} <- {}\n".format(item, result))
                return result
            meth = types.MethodType(my_func, self, Client)
        elif isinstance(func, Connection):
            sys.stderr.write("*-"*40+"\n")
            meth = func
        else:
            meth = func

        return meth

client = Client()

try:
    username, account, password = netrc.netrc().authenticators(VOSPACE_SERVER)
except:
    username = password = None

if username is None or password is None:
    username = raw_input('CANFAR Username: ')
    password = getpass('CANFAR Password: ')

authentication = HTTPBasicAuth(username, password)



__author__ = "David Rusk <drusk@uvic.ca>"

import os
import threading

from ossos import storage
from ossos.gui import logger


class SynchronizationManager(object):
    def __init__(self, remote_context, sync_enabled=False):
        self.remote_context = remote_context
        self.sync_enabled = sync_enabled

        self.syncable_files = []

    def enable_sync(self):
        self.sync_enabled = True
        self.sync_all()

    def disable_sync(self):
        self.sync_enabled = False

    def add_syncable_file(self, path):
        self.syncable_files.append(path)

        if self.sync_enabled:
            self.sync_all()

    def sync_all(self):
        while self.syncable_files:
            syncable_file = self.syncable_files.pop(0)
            thread = threading.Thread(target=self.do_synchronize,
                                      args=(syncable_file, ))
            thread.start()

    def do_synchronize(self, local_path):
        remote_uri = self.get_remote_uri(local_path)
        logger.info("Syncing %s to %s." % (local_path, remote_uri))
        storage.copy(local_path, remote_uri)

    def get_remote_uri(self, local_path):
        basename = os.path.basename(local_path)
        return self.remote_context.get_full_path(basename)




#!python
# A script to do validation
__author__ = "David Rusk <drusk@uvic.ca>"

import argparse
import tempfile

from ossos.gui import logger
from ossos.gui import tasks
from ossos import storage


def launch_app(task, working_directory, output_directory, dry_run, debug, name_filter=None,
               user_name=None, skip_previous=False, zoom=1):
    # Put import here to avoid the delay loading them.  This allows quick
    # feedback when argparse can tell the arguments are invalid, and makes
    # getting help with the -h flag faster.
    from ossos.gui.app import create_application

    create_application(task, working_directory, output_directory,
                       dry_run=dry_run, debug=debug, name_filter=name_filter,
                       user_id=user_name, skip_previous=skip_previous, zoom=zoom)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("task", choices=tasks.task_list,
                        help="The task to perform.")
    parser.add_argument("input",
                        help="The directory with files to be processed.")
    parser.add_argument("-o", "--output",
                        required=False,
                        default=tempfile.gettempdir(),
                        help="The directory where any local output files will "
                             "be placed.")
    parser.add_argument("--dbimages", default="vos:OSSOS/dbimages", help="VOSpace location of images to show.")
    parser.add_argument("--dry-run",
                        dest="dry_run",
                        action="store_true",
                        help="Do a dry run, where no results are sent to "
                             "VOSpace.")
    parser.add_argument("--debug",
                        action="store_true",
                        help="wx inspection tool will be launched.")
    parser.add_argument("--name-filter",
                        dest="name_filter",
                        help="A filter to apply to object names when loading from a directory.")
    parser.add_argument("--skip-previous", action="store_true",
                        dest="skip_previous",
                        help="Don't show me observation that are already in the input files, TRACKS only")
    parser.add_argument("--username", dest="username", help="Your CADC username")
    parser.add_argument('--zoom', dest='zoom', default=1)

    args = parser.parse_args()

    if args.dry_run and args.output == tempfile.gettempdir():
        # Don't use tempdir as default for dry runs, use the input directory
        output = args.input
    else:
        output = args.output

    if args.debug:
        logger.set_debug()

    storage.DBIMAGES = args.dbimages

    launch_app(args.task, args.input, output, args.dry_run, args.debug, args.name_filter,
               args.username, args.skip_previous, args.zoom)


if __name__ == "__main__":
    import warnings
    from astropy.utils.exceptions import AstropyUserWarning
    from astropy.io.fits.verify import VerifyWarning
    warnings.filterwarnings('ignore', category=AstropyUserWarning)
    warnings.filterwarnings('ignore', category=VerifyWarning)
    main()



from astropy.coordinates import SkyCoord
from astropy.io import fits

__author__ = "David Rusk <drusk@uvic.ca>"

import unittest
from astropy import units
from mock import patch
#from hamcrest import assert_that, equal_to
from astropy import table

from ossos import storage, mpc


class ConeSearchTest(unittest.TestCase):

    def test_cone_search(self):
        mpc_line="     O13AE3Y* C2013 04 04.42583 14 09 16.989-11 14 50.60         23.18r      568 1615909p27 O13AE3Y Y  1715.1 2159.5 23.18 0.10 UUUU %"
        observation = mpc.Observation.from_string(mpc_line)

        result_table = storage.cone_search(observation.coordinate.lonangle.degrees,
                            observation.coordinate.latangle.degrees,
                            )
        self.assertIsInstance(result_table, table.Table)
        self.assertEquals(result_table['dataset_name'][0],1607614)


class DownloadImages(unittest.TestCase):

    def test_get_cutout(self):
        uri = "vos:jkavelaars/2001_QT297/FORS2.2013-06-15T08:18:06.831/FORS2.2013-06-15T08:18:06.831.fits"
        ra = 340.20001892 * units.degree
        dec = -6.82226166 * units.degree
        coo = SkyCoord(ra, dec)
        radius = 10.0 * units.arcminute
        hdulist = storage.ra_dec_cutout(uri, coo, radius)
        self.assertIsInstance(hdulist, fits.HDUList)
        self.assertTrue(len(hdulist) > 1)
        for hdu in hdulist[1:]:
            self.assertTrue(hdu.header.get("XOFFSET", False) is not False)

class ObjectCountTest(unittest.TestCase):
    @patch("ossos.storage.set_property")
    @patch("ossos.storage.get_property")
    def test_increment_object_counter(self, get_property, set_property):
        get_property.return_value = "09"

        node_uri = "vos:drusk/OSSOS/astromdir/test"
        epoch_field = "13AE"

        counter = storage.increment_object_counter(node_uri, epoch_field)

        expected_tag = storage.build_counter_tag(epoch_field)
        expected_count = "0A"

        get_property.assert_called_once_with(node_uri, expected_tag, ossos_base=True)
        assert_that(counter, equal_to(expected_count))
        set_property.assert_called_once_with(node_uri, expected_tag,
                                             expected_count, ossos_base=True)

    @patch("ossos.storage.set_property")
    @patch("ossos.storage.get_property")
    def test_increment_counter_doesnt_exist(self, get_property, set_property):
        get_property.return_value = None

        node_uri = "vos:drusk/OSSOS/astromdir/test"
        epoch_field = "13AE"

        counter = storage.increment_object_counter(node_uri, epoch_field)

        expected_tag = storage.build_counter_tag(epoch_field)
        expected_count = "01"

        get_property.assert_called_once_with(node_uri, expected_tag, ossos_base=True)
        assert_that(counter, equal_to(expected_count))
        set_property.assert_called_once_with(node_uri, expected_tag,
                                             expected_count, ossos_base=True)

    def test_build_dryrun_tag(self):
        epoch_field = "13AE"

        assert_that(storage.build_counter_tag(epoch_field, dry_run=True),
                    equal_to("13AE-object_count-DRYRUN"))

    @patch("ossos.storage.set_property")
    @patch("ossos.storage.get_property")
    def test_increment_object_counter_dryrun(self, get_property, set_property):
        get_property.return_value = "02"

        node_uri = "vos:drusk/OSSOS/astromdir/test"
        epoch_field = "13AE"

        counter = storage.increment_object_counter(node_uri, epoch_field,
                                                   dry_run=True)

        expected_tag = storage.build_counter_tag(epoch_field, dry_run=True)
        expected_count = "03"

        assert_that(counter, equal_to(expected_count))
        get_property.assert_called_once_with(
            node_uri,
            storage.build_counter_tag(epoch_field, dry_run=True),
            ossos_base=True)
        set_property.assert_called_once_with(node_uri,
                                             expected_tag,
                                             expected_count,
                                             ossos_base=True)


if __name__ == '__main__':
    unittest.main()



# fields must be
#
# - within 25 degrees of opposition during New Moon in April
# - visible for +4 hours per night in April Dark Time
# - visible for +2 hours per night in Feb/Mar and May/Jun

import math
import sys

import ephem

cfht = ephem.Observer()
cfht.lat = 0.344
cfht.lon = -2.707
cfht.horizon = math.radians(20)
cfht.elevation = 3900
o = cfht
target_latitude = math.radians(0)
p = ephem.Ecliptic(0, target_latitude)

start_date = ephem.date('2015/06/01 00:00:00')
Months = ["", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
field = ephem.FixedBody()
Moon = ephem.Moon()
Sun = ephem.Sun()

if len(sys.argv) != 3:

    # for each day in the semester
    sys.stdout.write("%8s | %17s | %17s | %38s  %38s  %38s  \n" % (
        "Day", "%Ill Moon-2015 RA", "%Ill Moon-2016 RA", "Min-RA [uptime] (Glat)", "Opp-RA [uptime] (Glat)",
        "Max-RA [uptime] (Glat)"))

    for day in range(0, 360):
        o.date = ephem.date(start_date + day)
        Sun.compute(o)

        # Start at local Noon
        solar_noon = Sun.transit_time

        # Determine the next Sun Set and next Sun Rise (using dark time)
        o.horizon = "-18:00"
        o.date = solar_noon
        sun_set = o.next_setting(Sun)
        sun_rise = o.next_rising(Sun)

        # flip back to requiring source above airmass 1.5
        o.horizon = math.radians(20)

        Moon.compute(o)

        # Figure out where opposition is
        opp = ephem.Ecliptic(ephem.Sun(o))
        opp.lon += math.radians(180.0)
        opp.lat = target_latitude
        t = o.date.tuple()
        sys.stdout.write("%5s %2d | %3.0f %13s | " % (Months[t[1]], t[2], Moon.phase, Moon.ra))
        sys.stdout.write("%3.0f %13s |" % (ephem.Moon(o.date + 365).phase, ephem.Moon(o.date + 365).ra))

        field_centres = [(False, False, False),
                         (False, False, False),
                         (False, False, False)]

        prev_uptime = 0

        for ang in range(-20, 20, 1):
            o.horizon = math.radians(40)
            pos = ephem.Ecliptic(opp)
            pos.lon += math.radians(ang)
            pos.lat = target_latitude
            (field._ra, field._dec) = pos.to_radec()
            o.date = solar_noon
            field.compute(o)

            # we can start observing when the field is up and the sun is down.
            field_rising = o.next_rising(field)
            start_time = max(sun_set, field_rising)

            # next compute how long till the source sets or the sun rises again
            o.date = start_time
            field_setting = o.next_setting(field)
            end_time = min(field_setting, sun_rise)

            # also need to compute the angle between the moon and the field.
            uptime = (end_time - start_time) * 24.0

            # Now, if the moon is more than 25% illuminated we only observe when
            # the moon is down.
            o.horizon = 0.0
            Moon.compute(o)
            # remove from uptime the time the moon and the field where both up.
            both_up_start = max(Moon.rise_time, field_rising)
            both_up_end = min(Moon.set_time, field_setting)

            if Moon.phase > 25:
                if both_up_end > both_up_start:
                    uptime -= (both_up_end - both_up_start) * 24.0
            uptime = max(0, uptime)
            # print sun_set, field_rising, both_up_start, Moon.rise_time, Moon.set_time, both_up_end, field_setting,
            #  sun_rise, start_time, end_time, uptime
            # print o.date, o.next_setting(field), opp.lon, field.ra, field.dec, uptime
            if ang == 0 and uptime > 2:
                field_centres[1] = (field.ra, field.dec, uptime)
                prev_uptime = uptime
                continue
            if uptime < 2:
                continue
            if not field_centres[0][0]:
                field_centres[0] = (field.ra, field.dec, uptime)
                continue
            field_centres[2] = (field.ra, field.dec, uptime)

        o.horizon = math.radians(40)
        for p in field_centres:
            if not p:
                sys.stdout.write(" %13s %13s [%3.1f]" % (p[0], p[1], p[2]))
                field._ra = p[0]
                field._dec = p[1]
                field.compute(o)
                g = ephem.Galactic(field)
                sys.stdout.write(" (%3.0f)" % (math.degrees(g.lat)))
            else:
                sys.stdout.write("%40s" % "N/A")

        sys.stdout.write('\n')

else:

    start_date = ephem.date('2015/01/01 00:00:00')

    field._ra = ephem.hours(sys.argv[1])
    field._dec = ephem.degrees(sys.argv[2])
    field.compute(cfht)
    ec = ephem.Ecliptic(field)
    # ec.lat = target_latitude
    # (field._ra, field._dec) = ec.to_radec()
    # field.compute(cfht)

    for day in range(0, 180):
        o.date = ephem.date(start_date + day)
        Sun.compute(o)
        solar_noon = Sun.transit_time
        o.date = solar_noon
        o.horizon = math.radians(-18.0)
        sun_set = o.next_setting(Sun)
        sun_rise = o.next_rising(Sun)

        o.horizon = math.radians(40)
        field.compute(o)
        t = o.date.tuple()

        opp = ephem.Ecliptic(ephem.Sun(o))
        target = ephem.Ecliptic(field)
        opp.lat = target.lat
        elongation = math.degrees(target.lon - opp.lon)

        # we can start observing when the field is up and the sun is down.
        field_rising = o.next_rising(field)
        start_time = max(sun_set, field_rising)

        # next compute how long till the source sets or the sun rises again
        o.date = start_time
        field_setting = o.next_setting(field)
        end_time = min(field_setting, sun_rise)
        uptime = (end_time - start_time) * 24.0

        # also need to compute the angle between the moon and the field.
        # if the moon is more than 25% illuminated we only observe when
        # the moon is down.
        o.horizon = 0.0
        Moon.compute(o)
        # remove from uptime the time the moon and the field where both up.
        both_up_start = max(Moon.rise_time, field_rising)
        both_up_end = min(Moon.set_time, field_setting)

        if Moon.phase > 25:
            if both_up_end > both_up_start:
                uptime -= (both_up_end - both_up_start) * 24.0
        uptime = max(0, uptime)
        # print sun_set, field_rising, both_up_start, Moon.rise_time, Moon.set_time, both_up_end, field_setting,
        # sun_rise, start_time, end_time, uptime

        # Do the same computations for a year later.  For planning of recovery on this field.
        o.date += 365
        o.date = o.next_rising(field)
        Moon2 = ephem.Moon(o)
        uptime2 = (o.next_setting(field) - o.date) * 24

        sys.stdout.write("%5s %2d | %3.0f %13s [%3.1f] | " % (Months[t[1]], t[2], Moon.phase, Moon.ra, uptime))
        sys.stdout.write("%3.0f %13s |" % (Moon2.phase, Moon2.ra))
        sys.stdout.write(" %13s %13s |" % (field.ra, field.dec))
        sys.stdout.write(" %5.0f" % elongation)
        sys.stdout.write('\n')



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from mock import Mock
from hamcrest import assert_that, equal_to

from tests.base_tests import WxWidgetTestCase
from ossos.gui.controllers import ProcessRealsController
from ossos.gui.views.validation import SourceValidationPanel


class TestSourceValidationPanel(WxWidgetTestCase):
    def test_accept_button(self):
        validation_controller = Mock(spec=ProcessRealsController)
        validation_panel = SourceValidationPanel(self.rootframe, validation_controller)

        assert_that(validation_controller.on_accept.call_count, equal_to(0))

        accept_button = self.get_child_by_label(
            validation_panel, validation_panel.accept_label)
        self.fire_button_click_event(accept_button)

        assert_that(validation_controller.on_accept.call_count, equal_to(1))
        assert_that(validation_controller.on_reject.call_count, equal_to(0))

    def test_reject_button(self):
        validation_controller = Mock()
        validation_panel = SourceValidationPanel(self.rootframe, validation_controller)

        assert_that(validation_controller.on_reject.call_count, equal_to(0))

        reject_button = self.get_child_by_label(
            validation_panel, validation_panel.reject_label)
        self.fire_button_click_event(reject_button)

        assert_that(validation_controller.on_accept.call_count, equal_to(0))
        assert_that(validation_controller.on_reject.call_count, equal_to(1))


if __name__ == '__main__':
    unittest.main()



from astropy import units
from astropy.time import TimeDelta
from ...gui import events
from ...gui import logger, config
from ...gui.models.collections import StatefulCollection
from ...gui.models.exceptions import (ImageNotLoadedException,
                                      NoWorkUnitException,
                                      NoAvailableWorkException)
from ...gui.models.workload import (CandidatesWorkUnit, RealsWorkUnit,
                                    TracksWorkUnit)
from ...astrom import SourceReading
from ...downloads.cutouts.source import SourceCutout
from ...mpc import Time
__author__ = "David Rusk <drusk@uvic.ca>"


class ValidationModel(object):
    """
    Contains the data and associated operations available to the user interface.
    """

    def __init__(self, workunit_provider, image_manager, synchronization_manager):
        self.workunit_provider = workunit_provider
        self.image_manager = image_manager
        self.synchronization_manager = synchronization_manager

        self.work_units = StatefulCollection()

        self.num_processed = 0

        self.sources_discovered = set()

        self.image_state = SingletState(self)

    def get_working_directory(self):
        return self.workunit_provider.directory

    def start_work(self):
        logger.debug("Model starting work.")
        self.next_workunit()

    def next_source(self):
        self.get_current_workunit().next_source()
        self.expect_source_transition()

    def previous_source(self):
        self.get_current_workunit().previous_source()
        self.expect_source_transition()

    def next_obs(self):
        self.get_current_workunit().next_obs()
        self.expect_observation_transition()

    def previous_obs(self):
        self.get_current_workunit().previous_obs()
        self.expect_observation_transition()

    def next_item(self):
        if self.get_current_workunit().is_finished():
            self.next_workunit()
        else:
            self.get_current_workunit().next_item()

        if self.is_processing_candidates():
            self.expect_source_transition()
        elif self.is_processing_reals() or self.is_processing_tracks():
            self.expect_observation_transition()

    def accept_current_item(self):
        self.sources_discovered.add(self.get_current_source())
        self._process_current_item()

    def reject_current_item(self):
        self._process_current_item()

    def _process_current_item(self):
        self.get_current_workunit().process_current_item()
        self.num_processed += 1

    def next_workunit(self):
        if self.work_units.is_on_last_item():
            try:
                self._get_new_workunit()
            except NoAvailableWorkException:
                events.send(events.NO_AVAILABLE_WORK)
                return
        # # pre-load some work units if we are within 1 unit of the end
        # if self.work_units.index > len(self.work_units) - 1:
        #    try:
        #        self.workunit_provider.trigger_prefetching()
        #    except:
        #        pass
        self.work_units.next()

    def expect_source_transition(self):
        self.expect_image_transition()

    def expect_observation_transition(self):
        self.expect_image_transition()

    @staticmethod
    def expect_image_transition():
        events.send(events.CHANGE_IMAGE)

    def acknowledge_image_displayed(self):
        pass

    def add_workunit(self, new_workunit):
        new_workunit.register_finished_callback(self._on_finished_workunit)
        self.work_units.append(new_workunit)
        self.download_workunit_images(new_workunit)

    def _get_new_workunit(self):
        self.add_workunit(self.workunit_provider.get_workunit())

    def is_current_source_discovered(self):
        return self.get_current_source() in self.sources_discovered

    def is_current_item_processed(self):
        return self.get_current_workunit().is_current_item_processed()

    def is_current_source_finished(self):
        return self.get_current_workunit().is_current_source_finished()

    def is_current_source_adjusted(self):
        return self.get_current_cutout().is_adjusted()

    def get_num_items_processed(self):
        return self.num_processed

    def get_current_data(self):
        return self.get_current_workunit().get_data()

    def get_current_workunit(self):
        """
        @return the current unit of work.
        @rtype WorkUnit
        """
        workunit = self.work_units.get_current_item()
        if workunit is None:
            raise NoWorkUnitException()
        else:
            return workunit

    def set_current_workunit(self, workunit):
        self.work_units.set_current_item(workunit)

    def get_writer(self):
        return self.get_current_workunit().get_writer()

    def get_current_filename(self):
        return self.get_current_workunit().get_filename()

    def get_current_source_number(self):
        return self.get_current_workunit().get_current_source_number()

    def get_current_obs_number(self):
        return self.get_current_workunit().get_current_obs_number()

    def get_obs_count(self):
        return self.get_current_workunit().get_obs_count()

    def get_current_source(self):
        return self.get_current_workunit().get_current_source()

    def get_current_reading(self):
        """

        @return: the  SourceReading currently being validated.
        @rtype: SourceReading
        """
        return self.get_current_workunit().get_current_reading()

    def get_current_astrom_header(self):
        return self.get_current_reading().get_observation_header()

    def get_current_fits_header(self):
        return self.get_current_cutout().fits_header

    def get_current_exposure_number(self):
        return int(self.get_current_reading().obs.expnum)

    def get_reading_data(self):
        cutout = self.get_current_cutout()
        reading = self.get_current_reading()
        return (
            ("Original", "(x,y)"),
            (reading.x, reading.y),
            ("Reference", "(x,y)"),
            (reading.x0, reading.y0),
            ("Cutout", "(x,y)"),
            (cutout.pixel_x, cutout.pixel_y),
            ("R.A.", "DEC"),
            (reading.sky_coord.ra.to_string(unit='hour', sep=":", precision=2),
             reading.sky_coord.dec.to_string(unit='degree', sep=":", precision=1))
        )

    def get_header_data_list(self):

        try:
            header = self.get_current_astrom_header()
            header2 = self.get_current_fits_header()
            keys = ['MJD_OBS_CENTER',
                    'MJD-OBSC',
                    'EXPNUM',
                    'CHIPNUM', 'FWHM', 'FILTER']
            return [(key, header.get(key, header2.get(key, None))) for key in keys]
        except Exception as ex:
            print ex
            print "Failed to load mopheader, reverting to cutout header."
            pass

        try:
            header = self.get_current_cutout().hdulist[-1].header
            return [(key, value) for key, value in header.iteritems()]
        except Exception as ex:
            print ex
            pass

        return [('Header_Get', 'FAILED')]

    def get_current_observation_date(self):
        """
        Get the date of the current observation by looking in the header
        of the observation for the DATE and EXPTIME keywords.

        The 'DATE AT MIDDLE OF OBSERVATION' of the observation is returned
        @return: Time
        """
        # All HDU elements have the same date and time so just use
        # last one, sometimes the first one is missing the header, in MEF
        header = self.get_current_cutout().hdulist[-1].header
        mjd_obs = float(header.get('MJD-OBS'))
        exptime = float(header.get('EXPTIME'))
        mpc_date = Time(mjd_obs,
                        format='mjd',
                        scale='utc',
                        precision=config.read('MPC.DATE_PRECISION'))
        mpc_date += TimeDelta(exptime * units.second) / 2.0
        mpc_date = mpc_date.mpc
        return mpc_date

    def get_current_ra(self):
        try:
            return self.get_current_cutout().ra
        except ImageNotLoadedException:
            return self.get_current_reading().ra

    def get_current_dec(self):
        try:
            return self.get_current_cutout().dec
        except ImageNotLoadedException:
            return self.get_current_reading().dec

    def is_current_source_named(self):
        return self.get_current_source().has_provisional_name()

    def get_current_source_name(self):
        return self.get_current_source().get_provisional_name()

    def set_current_source_name(self, name):
        return self.get_current_source().set_provisional_name(name)

    def get_current_displayable_item(self):
        return self.image_state.get_current_displayable_item()

    def get_current_band(self):
        header = self.get_current_fits_header()
        filter_value = None
        for keyword in ['FILTER', 'FILT1 NAME']:
            filter_value = header.get(keyword, None)
            if filter_value is not None:
                break
        if filter_value is not None and filter_value.startswith('gri'):
            filter_value = 'w'
        return filter_value

    def get_current_pixel_source_point(self):
        return self.get_current_cutout().pixel_source_point

    def get_current_source_observed_magnitude(self):
        return self.get_current_cutout().get_observed_magnitude()

    def get_current_image_fwhm(self):
        return float(self.get_current_astrom_header().get("FWHM", 4.))

    def get_current_image_maxcount(self):
        return float(self.get_current_astrom_header()["MAXCOUNT"])

    def stop_loading_images(self):
        self.image_manager.stop_downloads()

    def start_loading_images(self):
        self.download_workunit_images(self.get_current_workunit())

    def submit_download_request(self, download_request):
        self.image_state.submit_download_request(download_request)

    def refresh_vos_client(self):
        self.image_manager.refresh_vos_clients()

    def update_current_source_location(self, new_location):
        raise NotImplementedError()

    def reset_current_source_location(self):
        raise NotImplementedError()

    def enable_synchronization(self):
        if self.synchronization_manager:
            self.synchronization_manager.enable_sync()
            logger.info("Synchronization enabled")

    def disable_synchronization(self):
        if self.synchronization_manager:
            self.synchronization_manager.disable_sync()
            logger.info("Synchronization disabled")

    def exit(self):
        for work_unit in self.work_units:
            work_unit.unlock()

        self.image_manager.stop_downloads()
        self.workunit_provider.shutdown()
        self.image_manager.wait_for_downloads_to_stop()

    def is_processing_candidates(self):
        return isinstance(self.get_current_workunit(), CandidatesWorkUnit)

    def is_processing_reals(self):
        return isinstance(self.get_current_workunit(), RealsWorkUnit)

    def is_processing_tracks(self):
        return isinstance(self.get_current_workunit(), TracksWorkUnit)

    def get_current_cutout(self):
        """

        @return: A SourceCutout of the currently active image.
        @rtype: SourceCutout
        """
        return self.image_state.get_current_cutout()

    def download_workunit_images(self, workunit):
        self.image_state.download_workunit_images(workunit)

    def use_singlets(self):
        logger.info("Model set to use image singlets.")
        self.image_state = SingletState(self)
        self.image_state.enter_state()

    def use_triplets(self):
        logger.info("Model set to use image triplets.")
        self.image_state = TripletState(self)
        self.image_state.enter_state()

    def _on_finished_workunit(self, results_file_paths):
        events.send(events.FINISHED_WORKUNIT, results_file_paths)

        if self.synchronization_manager:
            for path in results_file_paths:
                self.synchronization_manager.add_syncable_file(path)


class SingletState(object):
    def __init__(self, model):
        self.model = model
        self.image_manager = model.image_manager

    def enter_state(self):
        self.image_manager.stop_triplet_downloads()
        self.download_workunit_images(
            self.model.get_current_workunit())

    def get_current_cutout(self):
        return self.image_manager.get_cutout(
            self.get_current_displayable_item())

    def get_current_displayable_item(self):
        return self.model.get_current_reading()

    def download_workunit_images(self, workunit):
        self.image_manager.download_singlets_for_workunit(workunit)

    def submit_download_request(self, download_request):
        self.image_manager.submit_singlet_download_request(download_request)


class TripletState(object):
    def __init__(self, model):
        self.model = model
        self.image_manager = model.image_manager

    def enter_state(self):
        self.image_manager.stop_singlet_downloads()
        self.download_workunit_images(
            self.model.get_current_workunit())

    def get_current_cutout(self):
        return self.image_manager.get_cutout_grid(
            self.get_current_displayable_item())

    def get_current_displayable_item(self):
        return self.model.get_current_source()

    def download_workunit_images(self, workunit):
        self.image_manager.download_triplets_for_workunit(workunit)

    def submit_download_request(self, download_request):
        self.image_manager.submit_triplet_download_request(download_request)



import numpy

secrad = numpy.pi / 180.0 / 3600.0


def convert(lat, lon):
    """convert lat/lon from the ecliptic to the invariable plane."""

    x = numpy.cos(lon) * numpy.cos(lat)
    y = numpy.sin(lon) * numpy.cos(lat)
    z = numpy.sin(lat)

    # Invariable plane: values in arcseconds.
    epsilon = 5713.86
    omega = 387390.8

    coseps = numpy.cos(epsilon * secrad)
    sineps = numpy.sin(epsilon * secrad)
    cosom = numpy.cos(omega * secrad)
    sinom = numpy.sin(omega * secrad)

    xi = x * cosom + y * sinom
    yi = coseps * (-sinom * x + cosom * y) + sineps * z
    zi = - sineps * (-sinom * x + cosom * y) + coseps * z

    lat = numpy.arcsin(zi)
    lon = numpy.arctan2(yi, xi)
    return (lat, lon)


def trevonc(lat, lon, chiangchoi_plane=False):
    """convert lat/lon from the invariable to the ecliptic plane."""

    x = numpy.cos(lon) * numpy.cos(lat)
    y = numpy.sin(lon) * numpy.cos(lat)
    z = numpy.sin(lat)

    # default: the invariable plane. Values in arcseconds.
    # epsilon is the generalised descriptor of inclination
    # omega is the longitude of the ascending node
    epsilon = 5713.86
    omega = 387390.8
    if chiangchoi_plane is True:
        # an amplitude w.r.t. the ecliptic of 1.8 deg and nodal longitude of about 90 deg
        epsilon = 1.8 * 3600
        omega = 90. * 3600

    coseps = numpy.cos(epsilon * secrad)
    sineps = numpy.sin(epsilon * secrad)
    cosom = numpy.cos(omega * secrad)
    sinom = numpy.sin(omega * secrad)

    xi = cosom * x - sinom * (coseps * y - sineps * z)
    yi = sinom * x + cosom * (coseps * y - sineps * z)
    zi = sineps * y + coseps * z

    lat = numpy.arcsin(zi)
    lon = numpy.arctan2(yi, xi)
    return (lat, lon)



__author__ = "David Rusk <drusk@uvic.ca>"

ALPHABET_BASE_26 = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
ALPHABET_BASE_36 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
ALPHABET_BASE_52 = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
ALPHABET_BASE_62 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"


def base36encode(number, pad_length=0):
    return encode(number, ALPHABET_BASE_36).rjust(pad_length, "0")


def base36decode(number):
    return decode(number, 36)


def encode(number, alphabet):
    """
    Converts an integer to a base n string where n is the length of the
    provided alphabet.

    Modified from http://en.wikipedia.org/wiki/Base_36
    """
    if not isinstance(number, (int, long)):
        raise TypeError("Number must be an integer.")

    base_n = ""
    sign = ""

    if number < 0:
        sign = "-"
        number = -number

    if 0 <= number < len(alphabet):
        return sign + alphabet[number]

    while number != 0:
        number, i = divmod(number, len(alphabet))
        base_n = alphabet[i] + base_n

    return sign + base_n


def decode(number, base):
    return int(number, base)



__author__ = "David Rusk <drusk@uvic.ca>"

import wx


class WaitingGaugeDialog(wx.Dialog):

    def __init__(self, parent, wait_message, pulse_period_ms=100):
        super(WaitingGaugeDialog, self).__init__(parent)

        self.wait_message = wait_message
        self.pulse_period_ms = pulse_period_ms

        self._init_ui()

    def _init_ui(self):
        # Non-visible component used to periodically update gauge
        self.timer = wx.Timer(self)

        # Create visible components
        self.msg = wx.StaticText(self, label=self.wait_message)
        self.gauge = wx.Gauge(self)
        self.hidebutton = wx.Button(self, label="Hide")

        self._do_layout()

        self.hidebutton.Bind(wx.EVT_BUTTON, self._on_hide)
        self.Bind(wx.EVT_TIMER, self._on_tick, self.timer)

        self.timer.Start(self.pulse_period_ms)

    def _do_layout(self):
        vborder = 10
        vsizer = wx.BoxSizer(wx.VERTICAL)
        vsizer.Add(self.msg, flag=wx.CENTER | wx.TOP | wx.BOTTOM, border=vborder)
        vsizer.Add(self.gauge, flag=wx.EXPAND | wx.TOP | wx.BOTTOM, border=vborder)
        vsizer.Add(self.hidebutton, flag=wx.CENTER | wx.TOP | wx.BOTTOM, border=vborder)

        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        hsizer.Add(vsizer, flag=wx.LEFT | wx.RIGHT, border=20)

        self.SetSizer(hsizer)
        hsizer.Fit(self)

    def _on_tick(self, event):
        self.gauge.Pulse()

    def _on_hide(self, event):
        self.Hide()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from hamcrest import assert_that, equal_to, contains_inanyorder

from ossos.gui.context import VOSpaceWorkingContext

BASE_TEST_DIR = "vos:OSSOS/integration_tests/"
LISTING_TEST_DIR = BASE_TEST_DIR + "listing_test"


class VOSpaceWorkingContextTest(unittest.TestCase):
    def test_get_listing(self):
        context = VOSpaceWorkingContext(LISTING_TEST_DIR)

        assert_that(context.get_listing(".cands.astrom"),
                    contains_inanyorder("xxx1.cands.astrom", "xxx2.cands.astrom"))
        assert_that(context.get_listing(".reals.astrom"),
                    contains_inanyorder("xxx3.reals.astrom", "xxx4.reals.astrom"))

    def test_open(self):
        context = VOSpaceWorkingContext(BASE_TEST_DIR)

        filehandle = context.open("README.txt")
        file_contents = filehandle.read()
        filehandle.close()

        assert_that(file_contents,
                    equal_to("This directory contains files for OSSOS integration tests.\n"))

    def test_get_file_size(self):
        context = VOSpaceWorkingContext(BASE_TEST_DIR)

        assert_that(context.get_file_size("empty.txt"), equal_to(0))
        assert_that(context.get_file_size("README.txt"), equal_to(59))

    def test_exists(self):
        context = VOSpaceWorkingContext(BASE_TEST_DIR)

        assert_that(context.exists("README.txt"), equal_to(True))
        assert_that(context.exists("no_such_file.txt"), equal_to(False))


if __name__ == "__main__":
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

import wx

from mock import patch
from hamcrest import assert_that, equal_to

from ossos.gui.views.appview import guithread


class WxUtilTest(unittest.TestCase):
    @patch.object(wx, "CallAfter")
    @patch.object(wx, "Thread_IsMain")
    def test_threadsafe_from_worker_thread(
            self, Thread_IsMain_mock, CallAfter_mock):
        # Pretend we are not in the main GUI thread
        Thread_IsMain_mock.return_value = False

        @guithread
        def test_func():
            pass

        test_func()

        assert_that(Thread_IsMain_mock.call_count, equal_to(1))
        assert_that(CallAfter_mock.call_count, equal_to(1))

    @patch.object(wx, "CallAfter")
    @patch.object(wx, "Thread_IsMain")
    def test_threadsafe_from_main_thread(
            self, Thread_IsMain_mock, CallAfter_mock):
        # Pretend we ARE in the main GUI thread
        Thread_IsMain_mock.return_value = True

        @guithread
        def test_func():
            pass

        test_func()

        assert_that(Thread_IsMain_mock.call_count, equal_to(1))
        assert_that(CallAfter_mock.call_count, equal_to(0))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

import wx

from mock import Mock
from hamcrest import assert_that, equal_to

from tests.base_tests import WxWidgetTestCase
from ossos.gui.views.validation import KeyboardCompleteComboBox


class KeyboardCompleteComboBoxTest(WxWidgetTestCase):
    def setUp(self):
        super(KeyboardCompleteComboBoxTest, self).setUp()

        self.choice_default = ""
        self.choice_a = "A choice a"
        self.choice_b = "B choice b"
        self.choice_c = "C choice c"

        choices = [self.choice_default, self.choice_a, self.choice_b,
                   self.choice_c]

        self.undertest = KeyboardCompleteComboBox(self.rootframe, choices=choices)

        assert_that(self.undertest.GetValue(), equal_to(self.choice_default))

    def fire_mock_key_event(self, keycode):
        self.event = Mock(spec=wx.KeyEvent)
        self.event.GetKeyCode.return_value = keycode
        self.undertest._on_char(self.event)

    def get_selection(self):
        return self.undertest.GetStringSelection()

    def test_keyboard_completion(self):
        self.fire_mock_key_event(ord("B"))
        assert_that(self.get_selection(), equal_to(self.choice_b))

    def test_press_enter(self):
        self.fire_mock_key_event(wx.WXK_RETURN)
        self.event.Skip.assert_called_once_with()
        assert_that(self.get_selection(), equal_to(self.choice_default))
        self.fire_mock_key_event(ord("C"))
        assert_that(self.get_selection(), equal_to(self.choice_c))
        self.fire_mock_key_event(wx.WXK_RETURN)
        assert_that(self.get_selection(), equal_to(self.choice_c))
        self.event.Skip.assert_called_once_with()

    def test_invalid_character(self):
        self.fire_mock_key_event(wx.WXK_UP)
        assert_that(self.get_selection(), equal_to(self.choice_default))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"


class StatefulCollection(object):
    """
    An ordered collection of objects which have the notion of one of them
    being the 'current' object.
    """

    def __init__(self, items=None):
        if items is None:
            self.items = []
            self.index = -1
        else:
            self.items = items
            self.index = 0

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __getitem__(self, index):
        return self.items[index]

    def append(self, item):
        """Adds a new item to the end of the collection."""
        if len(self) == 0:
            # Special case, we make this the current item
            self.index = 0

        self.items.append(item)

    def insert(self, index, item):
        """
        Insert item at index location.
        """
        self.items.insert(index, item)

    def get_index(self):
        """Returns the index of the current item."""
        return self.index

    def set_current_item(self, item):
        """Repplace the current item with the provided item"""
        self.items[self.index] = item

    def get_current_item(self):
        if self.items:
            return self.items[self.index]
        else:
            return None

    def next(self):
        """
        Make the next item in the collection the current item.  Wraps around
        to the beginning after reaching the end.
        """
        self._move(1)

    def previous(self):
        """
        Make the previous item in the collection the current item.  Wraps
        around to the end after reaching the beginning.
        """
        self._move(-1)

    def is_on_last_item(self):
        """
        Returns True if the current item is the last item in the collection.
        """
        return self.index == len(self) - 1

    def _move(self, delta):
        self.index = (self.index + delta) % len(self)


from __future__ import absolute_import

import datetime
import os
import pprint
import warnings
import requests
import numpy

from astropy.io import ascii
from astropy import units
from astropy.coordinates import SkyCoord
from astropy.time import Time

from . import astrom, mpc, parameters
from .astrom import SourceReading
from .gui import logger
from .orbfit import Orbfit

requests.packages.urllib3.disable_warnings()

__author__ = 'Michele Bannister, JJ Kavelaars'

SSOS_URL = "http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/cadcbin/ssos/ssos.pl"
RESPONSE_FORMAT = 'tsv'
NEW_LINE = '\r\n'


class TracksParser(object):

    def __init__(self, inspect=True, skip_previous=False):
        logger.debug("Setting up TracksParser")
        self.orbit = None
        self._nights_per_darkrun = 18 * units.day
        self._nights_separating_darkruns = 30 * units.day
        self.inspect = inspect
        self.skip_previous = skip_previous
        self.ssos_parser = None

    def parse(self, filename):
        logger.debug("Parsing SSOS Query.")
        mpc_observations = mpc.MPCReader(filename).mpc_observations

        # pass down the provisional name so the table lines are linked to this TNO
        self.ssos_parser = SSOSParser(mpc_observations[0].provisional_name,
                                      input_observations=mpc_observations,
                                      skip_previous=self.skip_previous)

        try:
            self.orbit = Orbfit(mpc_observations)
            print self.orbit.residuals()
            print self.orbit
        except Exception as ex:
            logger.error("{}".format(ex))
            logger.error("Failed to compute orbit with astrometry provided")
            logger.error("{}".format(mpc_observations))
            return None

        if self.orbit.arc_length < 1 * units.day:
            # data from the same dark run.
            lunation_count = 0
        elif 1 * units.day < self.orbit.arc_length < self._nights_per_darkrun:
            # data from neighbouring darkruns.
            lunation_count = 1
        else:
            # data from the entire project.
            lunation_count = None

        # loop over the query until some new observations are found, or raise assert error.
        while True:
            tracks_data = self.query_ssos(mpc_observations, lunation_count)
            logger.debug("Got SSOS result: {}".format(tracks_data))
            if tracks_data.get_reading_count() > len(
                    mpc_observations) or tracks_data.get_arc_length() > self.orbit.arc_length + 2.0 * units.day:
                return tracks_data
            if not self.inspect:
                assert lunation_count is not None, "No new observations available."
            if lunation_count is None:
                return tracks_data
            lunation_count += 1
            if lunation_count > 2:
                lunation_count = None

    def query_ssos(self, mpc_observations, lunation_count=None):
        """Send a query to the SSOS web service, looking for available observations using the given track.

        :param mpc_observations: a list of mpc.Observations
        :param lunation_count: how many dark runs (+ and -) to search into
        :return: an SSOSData object
        :rtype: SSOSData
        """

        # we observe ~ a week either side of new moon
        # but we don't know when in the dark run the discovery happened
        # so be generous with the search boundaries, add extra 2 weeks
        # current date just has to be the night of the triplet,

        if lunation_count is None:
            # Only using SSOS to find data acquired during the survey period, for now.
            search_start_date = Time('2013-02-08', scale='utc')
            search_end_date = Time(datetime.datetime.now().strftime('%Y-%m-%d'), scale='utc')
        else:
            search_start_date = Time((mpc_observations[0].date.jd * units.day - (
                self._nights_per_darkrun +
                lunation_count * self._nights_separating_darkruns)), format='jd', scale='utc')
            search_end_date = Time((mpc_observations[-1].date.jd * units.day + (
                self._nights_per_darkrun +
                lunation_count * self._nights_separating_darkruns)), format='jd', scale='utc')

        logger.info("Sending query to SSOS start_date: {} end_data: {}\n".format(search_start_date, search_end_date))
        query = Query(mpc_observations,
                      search_start_date=search_start_date,
                      search_end_date=search_end_date)
        logger.debug("Parsing query results...")
        tracks_data = self.ssos_parser.parse(query.get(), mpc_observations=mpc_observations)

        tracks_data.mpc_observations = {}

        for mpc_observation in mpc_observations:
            # attach the input observations to the the SSOS query result.
            if isinstance(mpc_observation.comment, mpc.OSSOSComment):
                try:
                    tracks_data.mpc_observations[mpc_observation.comment.frame.strip()] = mpc_observation
                except Exception as e:
                    logger.error(str(e))
                    logger.error(mpc_observation)

        ref_sky_coord = None
        for source in tracks_data.get_sources():
            astrom_observations = tracks_data.observations
            source_readings = source.get_readings()
            foci = []
            for idx in range(len(source_readings)):
                source_reading = source_readings[idx]
                astrom_observation = astrom_observations[idx]
                self.orbit.predict(Time(astrom_observation.mjd, format='mjd', scale='utc'))
                assert isinstance(source_reading, SourceReading)
                if ref_sky_coord is None or source_reading.sky_coord.separation(ref_sky_coord) > 40 * units.arcsec:
                    foci.append([])
                    ref_sky_coord = source_reading.sky_coord
                foci[-1].append(source_reading)
            for focus in foci:
                ra = numpy.zeros(len(focus))
                dec = numpy.zeros(len(focus))
                for idx in range(len(focus)):
                    source_reading = focus[idx]
                    ra[idx] = source_reading.sky_coord.ra.to('degree').value
                    dec[idx] = source_reading.sky_coord.dec.to('degree').value
                ref_sky_coord = SkyCoord(ra.mean(), dec.mean(), unit='degree')
                for source_reading in focus:
                    source_reading.reference_sky_coord = ref_sky_coord
                    source_reading.pa = self.orbit.pa
                    # why are these being recorded just in pixels?  Because the error ellipse is drawn in pixels.
                    # TODO: Modify error ellipse drawing routine to use WCS but be sure
                    # that this does not cause trouble with the use of dra/ddec for cutout computer
                    source_reading.dx = self.orbit.dra
                    source_reading.dy = self.orbit.ddec
                    frame = astrom_observation.rawname
                    if frame in tracks_data.mpc_observations:
                        source_reading.discovery = tracks_data.mpc_observations[frame].discovery

        return tracks_data  # a SSOSData with .sources and .observations only


class TrackTarget(TracksParser):

    def parse(self, target_name):
        self.target_name = os.path.splitext(os.path.basename(target_name))[0]
        # pass down the provisional name so the table lines are linked to this TNO
        self.ssos_parser = SSOSParser(target_name,
                                      input_observations=None,
                                      skip_previous=False)

        return self.query_ssos(target_name=target_name)

    def query_ssos(self, target_name, lunation_count=None):
        """Send a query to the SSOS web service, looking for available observations using the given track.

        :param target_name: name of target to query against SSOIS db
        :param lunation_count: ignored
        :rtype: SSOSData
        """

        # we observe ~ a week either side of new moon
        # but we don't know when in the dark run the discovery happened
        # so be generous with the search boundaries, add extra 2 weeks
        # current date just has to be the night of the triplet,
        from . import horizons
        search_start_date = Time('2013-02-08', scale='utc')
        search_end_date = Time(datetime.datetime.now().strftime('%Y-%m-%d'), scale='utc')
        logger.info("Sending query to SSOS start_date: {} end_data: {}\n".format(search_start_date, search_end_date))
        query = Query(target_name,
                      search_start_date=search_start_date,
                      search_end_date=search_end_date)

        logger.debug("Parsing query results...")
        tracks_data = self.ssos_parser.parse(query.get())

        tracks_data.mpc_observations = {}

        start_time = Time(search_start_date)
        stop_time = Time(search_end_date)
        step_size = 5 * units.hour
        self.orbit = horizons.Body(target_name, start_time, stop_time, step_size)

        ref_sky_coord = None
        for source in tracks_data.get_sources():
            astrom_observations = tracks_data.observations
            source_readings = source.get_readings()
            for idx in range(len(source_readings)):
                source_reading = source_readings[idx]
                assert isinstance(source_reading, SourceReading)
                if ref_sky_coord is None or source_reading.sky_coord.separation(ref_sky_coord) > 40 * units.arcsec:
                    ref_sky_coord = source_reading.sky_coord
                source_reading.reference_sky_coord = ref_sky_coord
                astrom_observation = astrom_observations[idx]
                self.orbit.predict(Time(astrom_observation.mjd, format='mjd', scale='utc'))
                source_reading.pa = self.orbit.pa
                # why are these being recorded just in pixels?  Because the error ellipse is drawn in pixels.
                # TODO: Modify error ellipse drawing routine to use WCS but be sure
                # that this does not cause trouble with the use of dra/ddec for cutout computer
                source_reading.dx = self.orbit.dra
                source_reading.dy = self.orbit.ddec

        logger.debug("Sending back set of observations that might contain the target: {}".format(tracks_data))
        return tracks_data  # a SSOSData with .sources and .observations only


class SSOSParser(object):
    """
    Parse the result of an SSOS query, which is stored in an astropy Table object
    """

    def __init__(self, provisional_name, input_observations=None, skip_previous=False):
        """
        setup the parser.
        :param provisional_name: name of KBO to assign SSOS data to
        :param input_observations: input observations used in search
        """
        if input_observations is None:
            input_observations = []
        self.provisional_name = provisional_name
        self.input_rawnames = []
        self.null_observations = []
        self.skip_previous = skip_previous
        for observation in input_observations:
            if isinstance(observation.comment, mpc.OSSOSComment):
                try:
                    rawname = observation.comment.frame
                    self.input_rawnames.append(rawname.strip())
                    if observation.null_observation:
                        self.null_observations.append(rawname.strip())
                except Exception as ex:
                    logger.debug("{}".format(ex))
                    logger.debug("Failed to get original filename from >{}<".format(observation.comment))

    @staticmethod
    def _skip_missing_data(str_vals, ncols):
        """
        add a extra columna if missing, else return None.
        """
        while len(str_vals) < ncols:
            str_vals.append('None')
        return str_vals

    @staticmethod
    def build_source_reading(expnum, ccd=None, ftype='p'):
        """
        Build an astrom.Observation object for a SourceReading

        :param expnum: (str) Name or CFHT Exposure number of the observation.
        :param ccd: (str) CCD is this observation associated with. (can be None)
        :param ftype: (str) exposure time (specific to CFHT imaging)

        :return: An astrom.Observation object for the observation.
        :rtype: astrom.Observation
        """

        logger.debug("Building source reading for expnum:{} ccd:{} ftype:{}".format(expnum, ccd, ftype))

        return astrom.Observation(expnum=str(expnum),
                                  ftype=ftype,
                                  ccdnum=ccd)

    def parse(self, ssos_result_filename_or_lines, mpc_observations=None):
        """
        given the result table create 'source' objects.

        :param ssos_result_filename_or_lines:
        :param mpc_observations: a list of mpc.Observation objects used to retrieve the SSOS observations
        """
        table_reader = ascii.get_reader(Reader=ascii.Basic)
        table_reader.inconsistent_handler = self._skip_missing_data
        table_reader.header.splitter.delimiter = '\t'
        table_reader.data.splitter.delimiter = '\t'
        ssos_table = table_reader.read(ssos_result_filename_or_lines)

        sources = []
        observations = []
        source_readings = []

        if mpc_observations is not None and isinstance(mpc_observations[0], mpc.Observation):
            orbit = Orbfit(mpc_observations)
        else:
            from . import horizons
            start_time = Time(min(ssos_table['MJD']), format='mjd')
            stop_time = Time(max(ssos_table['MJD']), format='mjd')
            step_size = 5.0 * units.hour
            orbit = horizons.Body(self.provisional_name, start_time, stop_time, step_size)

        warnings.filterwarnings('ignore')
        logger.info("Loading {} observations\n".format(len(ssos_table)))
        expnums_examined = []
        for row in ssos_table:
            # Trim down to OSSOS-specific images
            if (row['Filter'] not in parameters.OSSOS_FILTERS) or row['Image_target'].startswith('WP'):
                continue

            # check if a dbimages object exists
            # For CFHT/MegaCam strip off the trailing character to get the exposure number.
            ftype = row['Image'][-1]
            expnum = row['Image'][:-1]

            # The file extension is the ccd number + 1 , or the first extension.
            ccd = int(row['Ext'])-1
            if 39 < ccd < 0:
                ccd = None
            x = row['X'] * units.pix
            y = row['Y'] * units.pix
            ra = row['Object_RA'] * units.degree
            dec = row['Object_Dec'] * units.degree
            ssois_coordinate = SkyCoord(ra, dec)
            mjd = row['MJD'] * units.day

            if not 0 < x.value < 2060 or not 0 < y.value < 4700:
                continue

            obs_date = Time(mjd, format='mjd', scale='utc')
            orbit.predict(obs_date)
            if orbit.dra > 4 * units.arcminute or orbit.ddec > 4.0 * units.arcminute:
                print "Skipping entry as orbit uncertainty at date {} is large.".format(obs_date)
                continue
            if expnum in expnums_examined:
                continue
            expnums_examined.append(expnum)

            logger.debug(("SSOIS Prediction: exposure:{} ext:{} "
                          "ra:{} dec:{} x:{} y:{}").format(expnum, ccd, ra, dec, x, y))

            logger.debug(("Orbfit Prediction: "
                          "ra:{} dec:{} ").format(orbit.coordinate.ra.to(units.degree),
                                                  orbit.coordinate.dec.to(units.degree)))

            observation = SSOSParser.build_source_reading(expnum, ccd, ftype=ftype)
            observation.mjd = mjd
            from_input_file = observation.rawname in self.input_rawnames
            # compare to input observation list.
            previous = False
            mpc_observation = None
            if from_input_file:
                for mpc_observation in mpc_observations:
                    try:
                        if mpc_observation.comment.frame.strip() == observation.rawname:
                            # only skip previous obseravtions if not discovery.
                            previous = not mpc_observation.discovery.is_initial_discovery
                            break
                    except Exception as e:
                        logger.debug(str(e))
                        pass
                    mpc_observation = None

            # skip previously measured observations if requested.
            if self.skip_previous and ( previous or observation.rawname in self.null_observations):
                continue

            logger.debug('built observation {}'.format(observation))
            observations.append(observation)
            null_observation = observation.rawname in self.null_observations

            ddec = orbit.ddec + abs(orbit.coordinate.dec - ssois_coordinate.dec)
            dra = orbit.dra + abs(orbit.coordinate.ra - ssois_coordinate.ra)

            logger.info(" Building SourceReading .... \n")
            source_reading = astrom.SourceReading(x=x, y=y, x0=x, y0=y,
                                                  ra=orbit.coordinate.ra.to(units.degree).value,
                                                  dec=orbit.coordinate.dec.to(units.degree).value,
                                                  xref=x, yref=y, obs=observation,
                                                  ssos=True, from_input_file=from_input_file,
                                                  dx=dra, dy=ddec, pa=orbit.pa,
                                                  null_observation=null_observation)
            source_reading.mpc_observation = mpc_observation
            source_readings.append(source_reading)

        # build our array of SourceReading objects
        sources.append(source_readings)

        warnings.filterwarnings('once')

        return SSOSData(observations, sources, self.provisional_name)


class SSOSData(object):
    """
    Encapsulates data extracted from an .astrom file.
    """

    def __init__(self, observations, sources, provisional_name):
        """
        Constructs a new astronomy data set object.

        Args:
          observations: list(Observations)
            The observations that are part of the data set.
        """
        self.mpc_observations = {}
        self.observations = observations
        self.sys_header = None
        self.sources = [astrom.Source(reading_list, provisional_name) for reading_list in sources]

    def get_reading_count(self):
        count = 0
        for source in self.sources:
            count += source.num_readings()

        return count

    def get_sources(self):
        return self.sources

    def get_source_count(self):
        return len(self.get_sources())

    def get_arc_length(self):
        mjds = []
        for obs in self.observations:
            mjds.append(obs.mjd)
        arc = (len(mjds) > 0 and max(mjds) - min(mjds)) or 0
        return arc


class ParamDictBuilder(object):
    """
    Build a dictionary of parameters needed for an SSOS Query.

    http://www4.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/cadcbin/ssos/ssosclf.pl?
    lang=en
    object=2
    search=bynameCADC
    epoch1=1990+01+01
    epoch2=2016+3+27
    eellipse=
    eunits=arcseconds
    extres=yes
    xyres=yes

    This should be fun!
    """

    def __init__(self,
                 observations=None,
                 verbose=False,
                 search_start_date=Time('2013-01-01', scale='utc'),
                 search_end_date=Time('2017-01-01', scale='utc'),
                 orbit_method='bern',
                 error_ellipse='bern',
                 resolve_extension=True,
                 resolve_position=True,
                 telescope_instrument='CFHT/MegaCam'):

        self._orbit_method = orbit_method
        self._verbose = False
        self.verbose = verbose
        self._search_start_date = None
        self.search_start_date = search_start_date
        self._search_end_date = None
        self.search_end_date = search_end_date
        self._error_ellipse = None
        self._resolve_extension = None
        self.resolve_extension = resolve_extension
        self._resolve_position = None
        self.resolve_position = resolve_position
        self._telescope_instrument = None
        self.telescope_instrument = telescope_instrument
        self._error_units = None
        self._observations = []
        self.observations = observations
        self.error_ellipse = error_ellipse

    @property
    def observations(self):
        """
        The observations to be used in fitting, returned as list of
        the mpc format lines.

        This should be set to a list of objects whose 'str' values
        will be valid MPC observations.
        """
        return self._observations

    @observations.setter
    def observations(self, observations):
        if not isinstance(observations, list) and not isinstance(observations, numpy.ndarray):
            observations = [observations]
        self._observations = []
        orbit_method_set = None
        for observation in observations:
            use_bern = isinstance(observation, mpc.Observation)
            self.orbit_method = use_bern and 'bern' or 'bynameCADC'
            orbit_method_set = orbit_method_set is None and self.orbit_method or orbit_method_set
            if orbit_method_set != self.orbit_method:
                raise ValueError("All members of observations list must be same type.")
            # use_bern needs to have any null observations removed.
            if use_bern and observation.null_observation:
                continue
            self._observations.append(str(observation))

    @property
    def verbose(self):
        """
        In verbose mode the SSOS query will return diagnostic
        information about how the search was done.
        """
        return self._verbose

    @verbose.setter
    def verbose(self, verbose):
        self._verbose = (verbose and 'yes') or 'no'

    @property
    def search_start_date(self):
        """
        :return: Time constraint for start of SSOS search window.
        :rtype: Time
        """
        return self._search_start_date

    @search_start_date.setter
    def search_start_date(self, search_start_date):
        """

        :param search_start_date: Time object for start of SSOS search window.
        """
        assert isinstance(search_start_date, Time)
        self._search_start_date = search_start_date.replicate(format='iso')
        self._search_start_date.out_subfmt = 'date'

    @property
    def search_end_date(self):
        """
        astropy.io.Time object. The end date of SSOS search window.
        """
        return self._search_end_date

    @search_end_date.setter
    def search_end_date(self, search_end_date):
        """
        :type search_end_date: astropy.io.Time
        :param search_end_date: search for frames take after the given date.
        """
        assert isinstance(search_end_date, Time)
        self._search_end_date = search_end_date.replicate(format='iso')
        self._search_end_date.out_subfmt = 'date'

    @property
    def orbit_method(self):
        """
        What fitting method should be used to turn the observations
        into an orbit.

        Must be one of ['bern', 'mpc']
        """
        return self._orbit_method

    @orbit_method.setter
    def orbit_method(self, orbit_method):
        assert orbit_method in ['bern', 'mpc', 'bynameCADC']
        self._orbit_method = orbit_method
        if self._orbit_method == 'bynameCADC':
            self.error_units = 'arcseconds'

    @property
    def error_units(self):
        return self._error_units

    @error_units.setter
    def error_units(self, error_units):
        self._error_units = error_units

    @property
    def error_ellipse(self):
        """
        The size of the error ellipse to assign to each position, or
        'bern' to use the output of the BK fit.
        """
        return self._error_ellipse

    @error_ellipse.setter
    def error_ellipse(self, error_ellipse):
        """

        :param error_ellipse: either a number or the work 'bern'
        """
        if not self.orbit_method == 'bern':
            try:
                error_ellipse = float(error_ellipse)
            except:
                error_ellipse = ''
        self._error_ellipse = error_ellipse

    @property
    def resolve_extension(self):
        """
        Should SSOS resolve and return which extension of a frame the
        object would be in?
        """
        return self._resolve_extension

    @resolve_extension.setter
    def resolve_extension(self, resolve_extension):
        if str(resolve_extension).lower() == "no":
            resolve_extension = False
        self._resolve_extension = (resolve_extension and "yes") or "no"

    @property
    def resolve_position(self):
        """
        Should SSOS resolve and return the predicted X/Y location of
        the source?
        """
        return self._resolve_position

    @resolve_position.setter
    def resolve_position(self, resolve_position):
        if str(resolve_position).lower() == "no":
            resolve_position = False
        self._resolve_position = (resolve_position and "yes") or "no"

    @property
    def telescope_instrument(self):
        """
        Name of the telescope being used.
        """
        return self._telescope_instrument

    @telescope_instrument.setter
    def telescope_instrument(self, telescope_instrument):
        assert isinstance(telescope_instrument, str)
        if telescope_instrument in TELINST:
            self._telescope_instrument = telescope_instrument

    @property
    def params(self):
        """
        :return: A dictionary of SSOS query parameters.
        :rtype: dict
        """
        params = dict(format=RESPONSE_FORMAT,
                      verbose=self.verbose,
                      epoch1=str(self.search_start_date),
                      epoch2=str(self.search_end_date),
                      search=self.orbit_method,
                      eunits=self.error_units,
                      eellipse=self.error_ellipse,
                      extres=self.resolve_extension,
                      xyres=self.resolve_position,
                      telinst=self.telescope_instrument)

        if self.orbit_method == 'bynameCADC':
            params['object'] = NEW_LINE.join((str(target_name) for target_name in self.observations))
        else:
            params['obs'] = NEW_LINE.join((str(observation) for observation in self.observations))
        return params


class Query(object):
    """
    Query the CADC's Solar System Object search for a given set of
    MPC-formatted moving object detection lines.

    Inputs:
        - a list of ossos.mpc.Observation instances

    Optional:
        - a tuple of the start and end times to be searched
          between. Format '%Y-%m-%d'

    Otherwise the temporal range defaults to spanning from the start
    of OSSOS surveying on 2013-01-01 to the present day.

    """

    def __init__(self,
                 observations=None,
                 search_start_date=Time(parameters.SURVEY_START, scale='utc'),
                 search_end_date=Time('2017-01-01', scale='utc'),
                 error_ellipse='bern'):

        self.param_dict_builder = ParamDictBuilder(
            observations=observations,
            search_start_date=search_start_date,
            search_end_date=search_end_date,
            error_ellipse=error_ellipse)
        self.headers = {'User-Agent': 'OSSOS'}

    def get(self):
        """
        :return: A string containing the TSV result from SSOS
        :rtype: str
        :raise: AssertionError
        """
        params = self.param_dict_builder.params
        logger.debug(pprint.pformat(format(params)))
        response = requests.post(SSOS_URL,
                                 data=params,
                                 headers=self.headers)
        logger.debug(response.url)
        assert isinstance(response, requests.Response)
        assert (response.status_code == requests.codes.ok)

        lines = response.content
        # note: spelling 'occured' is in SSOIS
        if len(lines) < 2 or "An error occured getting the ephemeris" in lines:
            print lines
            print response.url
            raise IOError(os.errno.EACCES,
                          "call to SSOIS failed on format error")

        if os.access("backdoor.tsv", os.R_OK):
            lines += open("backdoor.tsv").read()
        return lines

TELINST = [
    'AAT/WFI',
    'ALMA',
    'CFHT/CFH12K',
    'CFHT/MegaCam',
    'CFHT/WIRCam',
    'CTIO-4m/DECam',
    'CTIO-4m/Mosaic2',
    'CTIO-4m/NEWFIRM',
    'ESO-LaSilla_2.2m/WFI',
    'ESO-NTT/EFOSC',
    'ESO-NTT/EMMI',
    'ESO-NTT/SOFI',
    'ESO-NTT/SUSI',
    'ESO-NTT/SUSI2',
    'ESO-VISTA/VIRCAM',
    'ESO-VLT/FORS1',
    'ESO-VLT/FORS2',
    'ESO-VLT/HAWKI',
    'ESO-VLT/ISAAC',
    'ESO-VLT/NAOS-CONICA',
    'ESO-VLT/VIMOS',
    'ESO-VST/OMEGACAM',
    'Gemini/GMOS',
    'Gemini/GMOS-N',
    'Gemini/GMOS-S',
    'Gemini/NIRI',
    'HST/ACS',
    'HST/WFC3',
    'HST/WFPC2',
    'INT/WFC',
    'JKT/JAG-CCD',
    'KPNO-0.9m/Mosaic1',
    'KPNO-0.9m/Mosaic1.1',
    'KPNO-4m/Mosaic1',
    'KPNO-4m/Mosaic1.1',
    'KPNO-4m/NEWFIRM',
    'NEAT-GEODSS-Maui',
    'SDSS',
    'SOAR/SOI',
    'Subaru/SuprimeCam',
    'WHT/AUXCAM',
    'WHT/INGRID',
    'WHT/LIRIS',
    'WHT/Prime',
    'WISE',
    'WIYN/MiniMo',
    'WIYN/ODI',
]


__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from tests.base_tests import FileReadingTestCase
from ossos import daophot

DELTA = 0.0001


class DaophotTest(FileReadingTestCase):
    def test_phot(self):
        """
        Test data to compare with generated by running make testphot
        which runs the equivalent Perl script with the same data.
        """
        fits_filename = self.get_abs_path("data/1616681p22.fits")
        x_in = 560.06
        y_in = 406.51
        ap = 4
        insky = 11
        outsky = 15
        maxcount = 30000.0
        exptime = 1.0

        swidth = outsky - insky
        apcor = 0.0

        hdu = daophot.phot(fits_filename, x_in, y_in, aperture=ap, sky=insky,
                           swidth=swidth, apcor=apcor, maxcount=maxcount,
                           exptime=exptime)

        def get_first(param):
            value_list = hdu[param]
            self.assertEqual(len(value_list), 1)
            return value_list[0]

        xcen = get_first("XCENTER")
        ycen = get_first("YCENTER")
        mag = get_first("MAG")
        magerr = get_first("MERR")

        self.assertAlmostEqual(xcen, 560.0, 1)
        self.assertAlmostEqual(ycen, 406.6, 1)
        self.assertAlmostEqual(mag, 24.64, 2)
        self.assertAlmostEqual(magerr, 0.223, 2)

    def test_phot_mag(self):
        fits_filename = self.get_abs_path("data/1616681p22.fits")
        x_in = 560.06
        y_in = 406.51
        ap = 4
        insky = 11
        outsky = 15
        maxcount = 30000.0
        exptime = 1.0

        swidth = outsky - insky
        apcor = 0.0

        x, y, mag, magerr = daophot.phot_mag(fits_filename, x_in, y_in,
                                             aperture=ap, sky=insky,
                                             swidth=swidth, apcor=apcor,
                                             maxcount=maxcount,
                                             exptime=exptime)

        assert_that(x, close_to(560.000, DELTA))
        assert_that(y, close_to(406.600, DELTA))
        assert_that(mag, close_to(24.769, DELTA))
        # NOTE: minor difference in magnitude error: 0.290 vs 0.291
        assert_that(magerr, close_to(0.290, 0.0011))


if __name__ == '__main__':
    unittest.main()



__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from hamcrest import assert_that, equal_to
from mock import Mock

from tests.base_tests import FileReadingTestCase
from ossos.astrom import AstromParser
from ossos.downloads.cutouts.calculator import CoordinateConverter
from ossos.downloads.cutouts.source import SourceCutout


class SourceSnapshotTest(FileReadingTestCase):
    def setUp(self):
        test_file = "data/image_reading/realstest2.measure3.reals.astrom"
        astrom_data = AstromParser().parse(self.get_abs_path(test_file))

        self.reading = astrom_data.get_sources()[0].get_reading(0)
        self.original_observed_x = self.reading.x
        self.original_observed_y = self.reading.y

        self.offset_x = 10
        self.offset_y = 15
        coordinate_converter = CoordinateConverter(self.offset_x, self.offset_y)

        self.undertest = SourceCutout(self.reading, Mock())

        self.mock_update_ra_dec = Mock()
        self.undertest._update_ra_dec = self.mock_update_ra_dec

    def assert_update_ra_dec_call_count_equals(self, expected_calls):
        assert_that(self.mock_update_ra_dec.call_count,
                    equal_to(expected_calls))

    def test_xy_coordinate_systems(self):
        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y))

        expected_pixel_x = self.original_observed_x - self.offset_x
        expected_pixel_y = self.original_observed_y - self.offset_y

        assert_that(self.undertest.pixel_x, equal_to(expected_pixel_x))
        assert_that(self.undertest.pixel_y, equal_to(expected_pixel_y))

        assert_that(self.undertest.observed_source_point,
                    equal_to((self.original_observed_x, self.original_observed_y)))
        assert_that(self.undertest.pixel_source_point,
                    equal_to((expected_pixel_x, expected_pixel_y)))

    def test_xy_adjusted(self):
        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y))

        assert_that(self.undertest.is_adjusted(), equal_to(False))

        diff_x = 2.5
        diff_y = 3.5

        new_pixel_x = self.undertest.pixel_x + diff_x
        new_pixel_y = self.undertest.pixel_y + diff_y

        self.undertest.update_pixel_location((new_pixel_x, new_pixel_y))

        assert_that(self.undertest.pixel_x, equal_to(new_pixel_x))
        assert_that(self.undertest.pixel_y, equal_to(new_pixel_y))

        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x + diff_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y + diff_y))

        assert_that(self.undertest.is_adjusted(), equal_to(True))

    def test_update_xy_updates_ra_dec(self):
        self.assert_update_ra_dec_call_count_equals(0)

        self.undertest.update_pixel_location((665.0, 3215.0))

        _ = self.undertest.dec

        self.assert_update_ra_dec_call_count_equals(1)

    def test_ra_dec_updates_lazilly(self):
        self.assert_update_ra_dec_call_count_equals(0)

        self.undertest.update_pixel_location((665.0, 3215.0))

        self.assert_update_ra_dec_call_count_equals(0)

        self.undertest.update_pixel_location((665.0, 3218.0))

        self.assert_update_ra_dec_call_count_equals(0)

        _ = self.undertest.ra

        self.assert_update_ra_dec_call_count_equals(1)

        _ = self.undertest.dec

        self.assert_update_ra_dec_call_count_equals(1)

        self.undertest.update_pixel_location((667.2, 3216.6))

        self.assert_update_ra_dec_call_count_equals(1)

        _ = self.undertest.dec
        _ = self.undertest.ra

        self.assert_update_ra_dec_call_count_equals(2)

    def test_original_xy(self):
        assert_that(self.undertest.observed_x, equal_to(self.original_observed_x))
        assert_that(self.undertest.observed_y, equal_to(self.original_observed_y))

        assert_that(self.undertest.original_observed_x,
                    equal_to(self.original_observed_x))
        assert_that(self.undertest.original_observed_y,
                    equal_to(self.original_observed_y))

        diff_x = 2.5
        diff_y = 3.5

        new_pixel_x = self.undertest.pixel_x + diff_x
        new_pixel_y = self.undertest.pixel_y + diff_y

        self.undertest.update_pixel_location((new_pixel_x, new_pixel_y))

        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x + diff_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y + diff_y))

        assert_that(self.undertest.original_observed_x,
                    equal_to(self.original_observed_x))
        assert_that(self.undertest.original_observed_y,
                    equal_to(self.original_observed_y))

    def test_reset_location(self):
        original_pixel_x, original_pixel_y = self.undertest.pixel_source_point

        diff_x = 2.5
        diff_y = 3.5
        new_pixel_x = self.undertest.pixel_x + diff_x
        new_pixel_y = self.undertest.pixel_y + diff_y

        self.undertest.update_pixel_location((new_pixel_x, new_pixel_y))

        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x + diff_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y + diff_y))

        assert_that(self.undertest.pixel_x, equal_to(new_pixel_x))
        assert_that(self.undertest.pixel_y, equal_to(new_pixel_y))

        self.undertest.reset_source_location()

        assert_that(self.undertest.observed_x,
                    equal_to(self.original_observed_x))
        assert_that(self.undertest.observed_y,
                    equal_to(self.original_observed_y))

        assert_that(self.undertest.pixel_x, equal_to(original_pixel_x))
        assert_that(self.undertest.pixel_y, equal_to(original_pixel_y))

    def test_reset_location_is_adjusted_false(self):
        assert_that(self.undertest.is_adjusted(), equal_to(False))

        self.undertest.update_pixel_location((100, 100))

        assert_that(self.undertest.is_adjusted(), equal_to(True))

        self.undertest.reset_source_location()

        assert_that(self.undertest.is_adjusted(), equal_to(False))

    def test_reset_location_makes_data_stale(self):
        self.assert_update_ra_dec_call_count_equals(0)

        self.undertest.update_pixel_location((665.0, 3215.0))

        _ = self.undertest.ra

        self.assert_update_ra_dec_call_count_equals(1)

        self.undertest.reset_source_location()

        self.assert_update_ra_dec_call_count_equals(1)

        _ = self.undertest.ra

        self.assert_update_ra_dec_call_count_equals(2)


if __name__ == '__main__':
    unittest.main()



"""
This module contains various dialogs used in the application.
The accept/reject source dialogs are in validation.py.
"""
__author__ = "David Rusk <drusk@uvic.ca>"

import wx

from ossos.gui import config
from ossos.gui.views.listctrls import ListCtrlPanel


def should_exit_prompt(parent):
    dialog = wx.MessageDialog(parent,
                              config.read("UI.ALLPROC.MSG"),
                              caption=config.read("UI.ALLPROC.CAPTION"),
                              style=wx.YES_NO | wx.ICON_INFORMATION)

    user_choice = dialog.ShowModal()
    dialog.Destroy()

    return True if user_choice == wx.ID_YES else False


def show_empty_workload_dialog(parent, directory):
    message = ("No work to be done in %s\n"
               "It was either already processed or has no input files "
               "for the selected task." % directory)
    dialog = wx.MessageDialog(parent,
                              message,
                              caption="Empty Workload",
                              style=wx.OK | wx.ICON_INFORMATION)

    dialog.ShowModal()
    dialog.Destroy()


def show_keymappings_dialog(parent, keybind_manager):
    dialog = wx.Dialog(parent, title="Key Mappings")
    panel = ListCtrlPanel(dialog, ("Action", "Shortcut"))
    panel.populate_list(keybind_manager.get_keymappings())

    dialog.Show()


__author__ = "David Rusk <drusk@uvic.ca>"

import unittest

from mock import Mock
from hamcrest import (assert_that, contains_inanyorder, has_length, contains,
                      equal_to)

from tests.base_tests import FileReadingTestCase
from ossos.gui import tasks
from ossos.gui.context import LocalDirectoryWorkingContext
from ossos.gui.progress import (LocalProgressManager, InMemoryProgressManager,
                                   FileLockedException, RequiresLockException,
                                   LOCK_SUFFIX)

WD_HAS_PROGRESS = "data/persistence_has_progress"
WD_NO_LOG = "data/persistence_no_log"


class ProgressManagerLoadingTest(FileReadingTestCase):
    def setUp(self):
        self.working_directory = self.get_abs_path(WD_HAS_PROGRESS)
        directory_manager = LocalDirectoryWorkingContext(self.working_directory)
        self.progress_manager = LocalProgressManager(directory_manager)

    def tearDown(self):
        self.progress_manager.clean(suffixes=[LOCK_SUFFIX])

    def test_load_progress(self):
        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    contains_inanyorder("xxx1.cands.astrom", "xxx3.cands.astrom"))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    contains_inanyorder("xxx3.reals.astrom"))


class ProgressManagerFreshDirectoryTest(FileReadingTestCase):
    def setUp(self):
        self.main_user_id = "main_user"
        self.working_directory = self.get_abs_path(WD_NO_LOG)
        directory_manager = LocalDirectoryWorkingContext(self.working_directory)
        self.progress_manager = LocalProgressManager(directory_manager,
                                                     userid=self.main_user_id)

    def create_concurrent_progress_manager(self, userid="test_user"):
        directory_manager = LocalDirectoryWorkingContext(self.working_directory)
        return LocalProgressManager(directory_manager, userid=userid)

    def tearDown(self):
        # Get rid of generated files so we don't interfere with other tests
        self.progress_manager.clean()

    def test_load_progress_no_logs(self):
        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    has_length(0))

    def test_record_done_requires_lock(self):
        self.assertRaises(
            RequiresLockException,
            self.progress_manager.record_done, "xxx2.reals.astrom")

    def test_write_progress(self):
        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    has_length(0))

        processed1 = "xxx2.reals.astrom"
        self.progress_manager.lock(processed1)
        self.progress_manager.record_done(processed1)

        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    contains(processed1))
        assert_that(self.progress_manager.is_done(processed1), equal_to(True))

        # Create a second persistence manager and make sure it sees the changes
        manager2 = self.create_concurrent_progress_manager()
        assert_that(manager2.get_done(tasks.CANDS_TASK), has_length(0))
        assert_that(manager2.get_done(tasks.REALS_TASK), contains(processed1))
        assert_that(manager2.is_done(processed1), equal_to(True))

    def test_write_progress_two_simultaneous_managers(self):
        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    has_length(0))

        processed1 = "xxx2.reals.astrom"
        self.progress_manager.lock(processed1)
        self.progress_manager.record_done(processed1)

        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    contains(processed1))

        # Create a second simultaneous manager
        manager2 = self.create_concurrent_progress_manager()
        processed2 = "xxx3.reals.astrom"
        manager2.lock(processed2)
        manager2.record_done(processed2)

        # Make sure second manager sees both entries
        assert_that(manager2.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(manager2.get_done(tasks.REALS_TASK),
                    contains_inanyorder(processed1, processed2))

        # Make sure original manager sees both entries
        assert_that(self.progress_manager.get_done(tasks.CANDS_TASK),
                    has_length(0))
        assert_that(self.progress_manager.get_done(tasks.REALS_TASK),
                    contains_inanyorder(processed1, processed2))

    def test_lock_file(self):
        file1 = "xxx1.cands.astrom"
        self.progress_manager.lock(file1)

        # No-one else should be able to acquire the lock...
        lock_requesting_user = "lock_requesting_user"
        manager2 = self.create_concurrent_progress_manager(lock_requesting_user)
        self.assertRaises(FileLockedException, manager2.lock, file1)

        # ... until we unlock it
        self.progress_manager.unlock(file1)
        manager2.lock(file1)

    def test_lock_has_locker_id(self):
        lock_holding_user = self.main_user_id
        lock_requesting_user = "lock_requesting_user"

        file1 = "xxx1.cands.astrom"
        self.progress_manager.lock(file1)

        manager2 = self.create_concurrent_progress_manager(lock_requesting_user)

        try:
            manager2.lock(file1)
            self.fail("Should have thrown FileLockedExcecption")
        except FileLockedException as ex:
            assert_that(ex.filename, equal_to(file1))
            assert_that(ex.locker, equal_to(lock_holding_user))

    def test_record_index_requires_lock(self):
        self.assertRaises(RequiresLockException,
                          self.progress_manager.record_index,
                          "xxx1.cands.astrom", 0)

    def test_record_index(self):
        file1 = "xxx1.cands.astrom"
        self.progress_manager.lock(file1)
        self.progress_manager.record_index(file1, 1)
        self.progress_manager.record_index(file1, 3)
        self.progress_manager.record_index(file1, 0)

        assert_that(self.progress_manager.get_processed_indices(file1),
                    contains_inanyorder(1, 3, 0))

        # Check they are still recorded after we release lock
        self.progress_manager.unlock(file1)
        assert_that(self.progress_manager.get_processed_indices(file1),
                    contains_inanyorder(1, 3, 0))

        # Check other clients can read them
        manager2 = self.create_concurrent_progress_manager()
        assert_that(manager2.get_processed_indices(file1),
                    contains_inanyorder(1, 3, 0))

    def test_unlock_after_record_done_no_error(self):
        file1 = "xxx1.cands.astrom"
        self.progress_manager.lock(file1)
        self.progress_manager.record_done(file1)
        self.progress_manager.unlock(file1)

    def test_record_done_does_not_unlock_all(self):
        file1 = "xxx1.cands.astrom"
        file2 = "xxx2.cands.astrom"
        manager2 = self.create_concurrent_progress_manager()

        self.progress_manager.lock(file1)
        manager2.lock(file2)

        self.progress_manager.record_done(file1)
        assert_that(manager2.owns_lock(file2), equal_to(True))

        manager2.unlock(file2)
        assert_that(manager2.owns_lock(file2), equal_to(False))

    def test_get_processed_indices_empty_should_not_cause_error(self):
        assert_that(self.progress_manager.get_processed_indices("xxx1.cands.astrom"),
                    has_length(0))

    def test_get_processed_indices_after_done(self):
        filename = "xxx1.cands.astrom"
        self.progress_manager.lock(filename)
        self.progress_manager.record_index(filename, 0)
        self.progress_manager.record_index(filename, 1)
        self.progress_manager.record_index(filename, 2)
        self.progress_manager.unlock(filename)

        assert_that(self.progress_manager.get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))

        self.progress_manager.lock(filename)
        self.progress_manager.record_done(filename)
        self.progress_manager.unlock(filename)

        assert_that(self.progress_manager.get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))

        # Double check with a second manager
        assert_that(self.create_concurrent_progress_manager().get_processed_indices(filename),
                    contains_inanyorder(0, 1, 2))


class InMemoryProgressManagerTest(unittest.TestCase):
    def setUp(self):
        self.file1 = "file1"
        self.file2 = "file2"

        directory_manager = Mock(spec=LocalDirectoryWorkingContext)
        self.undertest = InMemoryProgressManager(directory_manager)

    def test_done(self):
        assert_that(self.undertest.is_done(self.file1), equal_to(False))
        assert_that(self.undertest.is_done(self.file2), equal_to(False))

        self.undertest.lock(self.file2)
        self.undertest.record_done(self.file2)
        self.undertest.unlock(self.file2)

        assert_that(self.undertest.is_done(self.file1), equal_to(False))
        assert_that(self.undertest.is_done(self.file2), equal_to(True))

        self.undertest.lock(self.file1)
        self.undertest.record_done(self.file1)
        self.undertest.unlock(self.file1)

        assert_that(self.undertest.is_done(self.file1), equal_to(True))
        assert_that(self.undertest.is_done(self.file2), equal_to(True))

    def test_processed_indices(self):
        assert_that(self.undertest.get_processed_indices(self.file1),
                    has_length(0))
        self.undertest.lock(self.file1)
        self.undertest.record_index(self.file1, 1)
        assert_that(self.undertest.get_processed_indices(self.file1),
                    contains_inanyorder(1))

        self.undertest.record_index(self.file1, 2)
        assert_that(self.undertest.get_processed_indices(self.file1),
                    contains_inanyorder(1, 2))

        assert_that(self.undertest.get_processed_indices(self.file2),
                    has_length(0))

    def test_locking(self):
        assert_that(self.undertest.owns_lock(self.file1), equal_to(False))
        assert_that(self.undertest.owns_lock(self.file2), equal_to(False))

        self.undertest.lock(self.file1)

        assert_that(self.undertest.owns_lock(self.file1), equal_to(True))
        assert_that(self.undertest.owns_lock(self.file2), equal_to(False))

        self.assertRaises(RequiresLockException, self.undertest.record_done, self.file2)
        self.assertRaises(RequiresLockException, self.undertest.record_index, self.file2, 1)

    def test_external_locks(self):
        self.undertest.add_external_lock(self.file1)

        assert_that(self.undertest.owns_lock(self.file1), equal_to(False))
        self.assertRaises(FileLockedException, self.undertest.lock, self.file1)
        self.assertRaises(FileLockedException, self.undertest.unlock, self.file1)

        self.undertest.lock(self.file2)
        assert_that(self.undertest.owns_lock(self.file2), equal_to(True))


if __name__ == '__main__':
    unittest.main()
